# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import time, os, sys, numpy as np, shutil as sh
import glob
from mpl_toolkits.axes_grid1 import make_axes_locatable
from collections import OrderedDict
import json, yaml
import traceback

pjoin = os.path.join


def mkdirs(*paths):
    for p in paths:
        if not os.path.exists(p):
            os.makedirs(p)


def get_project_path(ExpID):
    full_path = glob.glob('Experiments/*%s*' % ExpID)
    assert (len(full_path) == 1
            )  # There should be only ONE folder with <ExpID> in its name.
    return full_path[0]


class LogPrinter(object):

    def __init__(self, file, ExpID, print_to_screen=False):
        self.file = file
        self.ExpID = ExpID
        self.print_to_screen = print_to_screen

    def __call__(self, *in_str):
        in_str = [str(x) for x in in_str]
        in_str = ' '.join(in_str)
        short_exp_id = self.ExpID[-6:]
        pid = os.getpid()
        current_time = time.strftime('%Y/%m/%d-%H:%M:%S')
        out_str = '[%s %s %s] %s' % (short_exp_id, pid, current_time, in_str)
        print(out_str, file=self.file, flush=True)  # print to txt
        if self.print_to_screen:
            print(out_str)  # print to screen

    def logprint(self, *in_str):  # to keep the interface uniform
        self.__call__(*in_str)

    def accprint(self, *in_str):
        blank = '  ' * int(self.ExpID[-1])
        self.__call__(blank, *in_str)

    def netprint(self, *in_str):  # i.e., print without any prefix
        """Deprecated.

        Use netprint in Logger.
        """
        for x in in_str:
            print(x, file=self.file, flush=True)
            if self.print_to_screen:
                print(x)

    def print(self, *in_str):
        """Print without any prefix."""
        for x in in_str:
            print(x, file=self.file, flush=True)
            if self.print_to_screen:
                print(x)

    def print_args(self, args):
        """
        Example: ('batch_size', 16) ('CodeID', 12defsd2) ('decoder', models/small16x_ae_base/d5_base.pth)
        It will sort the arg keys in alphabeta order, ignoring the upper/lower difference.
        """
        # build a key map for later sorting
        key_map = {}
        for k in args.__dict__:
            k_lower = k.lower()
            if k_lower in key_map:
                key_map[k_lower + '_' + k_lower] = k
            else:
                key_map[k_lower] = k

        # print in the order of sorted lower keys
        logtmp = ''
        for k_ in sorted(key_map.keys()):
            real_key = key_map[k_]
            logtmp += "('%s': %s) " % (real_key, args.__dict__[real_key])
        self.print(logtmp[:-1] + '\n')  # the last one is blank


class LogTracker(object):

    def __init__(self, momentum=0.9):
        self.loss = OrderedDict()
        self.momentum = momentum
        self.show = OrderedDict()

    def __call__(self, name, value, step=-1, show=True):
        """Update the loss value of <name>"""
        assert type(step) == int
        # value = np.array(value)

        if step == -1:
            if name not in self.loss:
                self.loss[name] = value
            else:
                self.loss[name] = self.loss[name] * self.momentum + value * (
                    1 - self.momentum)
        else:
            if name not in self.loss:
                self.loss[name] = [[step, value]]
            else:
                self.loss[name].append([step, value])

        # if the loss item will show in the log printing
        self.show[name] = show

    def avg(self, name):
        nparray = np.array(self.loss[name])
        return np.mean(nparray[:, 1], axis=0)

    def max(self, name):
        nparray = np.array(self.loss[name])
        # TODO: max index
        return np.max(nparray[:, 1], axis=0)

    def format(self):
        """
        loss example:
            [[1, xx], [2, yy], ...] ==> [[step, [xx, yy]], ...]
            xx ==> [xx, yy, ...]
        """
        keys = self.loss.keys()
        k_str, v_str = [], []
        for k in keys:
            if self.show[k] == False:
                continue
            v = self.loss[k]
            if not hasattr(v, '__len__'):  # xx
                v = '%.4f' % v
            else:
                if not hasattr(v[0], '__len__'):  # [xx, yy, ...]
                    v = ' '.join(['%.3f' % x for x in v])
                elif hasattr(v[0][1], '__len__'):  # [[step, [xx, yy]], ...]
                    v = ' '.join(['%.3f' % x for x in v[-1][1]])
                else:  # [[1, xx], [2, yy], ...]
                    v = '%.4f' % v[-1][1]

            length = min(max(len(k), len(v)), 15)
            format_str = '{:<%d}' % (length)
            k_str.append(format_str.format(k))
            v_str.append(format_str.format(v))
        k_str = ' | '.join(k_str)
        v_str = ' | '.join(v_str)
        return k_str + ' |', v_str + ' |'

    def plot(self, name, out_path):
        """Plot the loss of <name>, save it to <out_path>."""
        v = self.loss[name]
        if (not hasattr(v, '__len__')) or type(
                v[0][0]) != int:  # do not log the 'step'
            return
        if hasattr(v[0][1], '__len__'):
            # self.plot_heatmap(name, out_path)
            return
        v = np.array(v)
        step, value = v[:, 0], v[:, 1]
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(step, value)
        ax.set_xlabel('step')
        ax.set_ylabel(name)
        ax.grid()
        fig.savefig(out_path, dpi=200)
        plt.close(fig)

    def plot_heatmap(self, name, out_path, show_ticks=False):
        """
        A typical case: plot the training process of 10 weights
        x-axis: step
        y-axis: index (10 weights, 0-9)
        value: the weight values
        """
        v = self.loss[name]
        step, value = [], []
        [(step.append(x[0]), value.append(x[1])) for x in v]
        n_class = len(value[0])
        fig, ax = plt.subplots(figsize=[0.1 * len(step),
                                        n_class / 5])  # /5 is set manually
        im = ax.imshow(np.transpose(value), cmap='jet')

        # make a beautiful colorbar
        divider = make_axes_locatable(ax)
        cax = divider.append_axes('right', size=0.05, pad=0.05)
        fig.colorbar(im, cax=cax, orientation='vertical')

        # set the x and y ticks
        # For now, this can not adjust its range adaptively, so deprecated.
        # ax.set_xticks(range(len(step))); ax.set_xticklabels(step)
        # ax.set_yticks(range(len(value[0]))); ax.set_yticklabels(range(len(value[0])))

        interval = step[0] if len(step) == 1 else step[1] - step[0]
        ax.set_xlabel('step (* interval = %d)' % interval)
        ax.set_ylabel('index')
        ax.set_title(name)
        fig.savefig(out_path, dpi=200)
        plt.close(fig)


class Logger(object):
    """The top logger, which (1) set up all log directories (2) maintain the
    losses and accuracies."""

    def __init__(self, args):
        self.args = args

        # set up work folder
        self.ExpID = self.get_ExpID()
        self.Exps_Dir = 'Experiments'
        if hasattr(self.args, 'Exps_Dir'):
            self.Exps_Dir = self.args.Exps_Dir
        self.set_up_dir()

        self.log_printer = LogPrinter(
            self.logtxt, self.ExpID, self.args.debug
            or self.args.screen_print)  # for all txt logging
        self.log_tracker = LogTracker()  # for all numerical logging

        # initial print: save args
        self.print_script()
        self.print_nvidia_smi()
        self.print_note()
        if (not args.debug) and self.SERVER != '':
            # If self.SERVER != '', it shows this is Huan's computer, then call this func, which is just a small feature to my need.
            # When others use this code, they probably need NOT call this func.
            # self.__send_to_exp_hub() # this function is not very useful. deprecated.
            pass
        args.CodeID = self.get_CodeID()
        self.log_printer.print_args(args)
        self.save_args(args)
        self.cache_model()
        self.n_log_item = 0

    def get_CodeID(self):
        if hasattr(self.args, 'CodeID') and self.args.CodeID:
            return self.args.CodeID
        else:
            f = 'wh_git_status_%s.tmp' % time.time()
            script = 'git status >> %s' % f
            os.system(script)
            x = open(f).readlines()
            x = ''.join(x)
            os.remove(f)
            if 'Changes not staged for commit' in x:
                self.log_printer(
                    'Warning! Your code is not committed. Cannot be too careful.'
                )
                time.sleep(3)

            f = 'wh_CodeID_file_%s.tmp' % time.time()
            script = 'git log --pretty=oneline >> %s' % f
            os.system(script)
            x = open(f).readline()
            os.remove(f)
            return x[:8]

    def get_ExpID(self):
        self.SERVER = os.environ['SERVER'] if 'SERVER' in os.environ.keys(
        ) else ''
        TimeID = time.strftime('%Y%m%d-%H%M%S')
        ExpID = 'SERVER' + self.SERVER + '-' + TimeID
        return ExpID

    def set_up_dir(self):
        project_path = pjoin(
            '%s/%s_%s' % (self.Exps_Dir, self.args.project_name, self.ExpID))
        if hasattr(self.args, 'resume_ExpID') and self.args.resume_ExpID:
            project_path = get_project_path(self.args.resume_ExpID)
        if (
                self.args.debug
        ):  # debug has the highest priority. If debug, all the things will be saved in Debug_dir
            project_path = 'Debug_Dir'

        self.exp_path = project_path
        self.weights_path = pjoin(project_path, 'weights')
        self.gen_img_path = pjoin(project_path, 'gen_img')
        self.cache_path = pjoin(project_path, '.caches')
        self.log_path = pjoin(project_path, 'log')
        self.logplt_path = pjoin(project_path, 'log', 'plot')
        self.logtxt_path = pjoin(project_path, 'log', 'log.txt')
        mkdirs(self.weights_path, self.gen_img_path, self.logplt_path,
               self.cache_path)
        self.logtxt = open(self.logtxt_path, 'a+')
        self.script_hist = open(
            '.script_history',
            'a+')  # save local script history, for convenience of check

    def print_script(self):
        script = 'cd %s\n' % os.path.abspath(os.getcwd())
        if 'CUDA_VISIBLE_DEVICES' in os.environ:
            gpu_id = os.environ['CUDA_VISIBLE_DEVICES']
            script += ' '.join(
                ['CUDA_VISIBLE_DEVICES=%s python' % gpu_id, *sys.argv])
        else:
            script += ' '.join(['python', *sys.argv])
        script += '\n'
        print(script, file=self.logtxt, flush=True)
        print(script, file=sys.stdout, flush=True)
        print(script, file=self.script_hist, flush=True)

    def print_exc(self):
        traceback.print_exc(file=self.logtxt)

    def print_nvidia_smi(self):
        out = pjoin(self.log_path, 'gpu_info.txt')
        script = 'nvidia-smi >> %s' % out
        os.system(script)

    def print_note(self):
        project = self.get_project_name()  # the current project folder name
        exp_id = self.ExpID.split('-')[-1]  # SERVER138-20200623-095526
        if hasattr(self.args, 'note') and self.args.note:
            self.ExpNote = 'ExpNote [%s-%s-%s]: "%s" -- %s' % (
                self.SERVER,
                project,
                exp_id,
                self.args.note,
                self.args.project_name,
            )
            print(self.ExpNote, file=self.logtxt, flush=True)
            print(self.ExpNote, file=sys.stdout, flush=True)

    def plot(self, name, out_path):
        self.log_tracker.plot(name, out_path)

    def print(self, step):
        keys, values = self.log_tracker.format()
        k = keys.split('|')[0].strip()
        if k:  # only when there is sth to print, print
            values += ' (step = %d)' % step
            if (step % (self.args.print_interval * 10) == 0
                    or len(self.log_tracker.loss.keys()) > self.n_log_item
                ):  # when a new loss is added into the loss pool, print
                self.log_printer(keys)
                self.n_log_item = len(self.log_tracker.loss.keys())
            self.log_printer(values)

    def cache_model(self):
        """Save the model architecture, loss, configs, in case of future
        check."""
        extensions = [
            '.py',
            '.json',
            '.yaml',
            '.sh',
            '.txt',
            '.md',
        ]  # files of these types will be cached
        self.log_printer("==> Caching various config files to '%s'" %
                         self.cache_path)
        for root, dirs, files in os.walk('.'):
            if 'Experiments' in root or 'Debug_Dir' in root:
                continue
            for f in files:
                _, ext = os.path.splitext(f)
                if ext in extensions:
                    dir_path = pjoin(self.cache_path, root)
                    f_path = pjoin(root, f)
                    if not os.path.exists(dir_path):
                        os.makedirs(dir_path)
                    if os.path.exists(f_path):
                        sh.copy(f_path, dir_path)

    def get_project_name(self):
        """For example, 'Projects/CRD/logger.py', then return CRD."""
        file_path = os.path.abspath(__file__)
        return file_path.split('/')[-2]

    def __send_to_exp_hub(self):
        """For every experiment, it will send <ExpNote> to a hub for the
        convenience of checking."""
        today_exp = time.strftime('%Y%m%d') + '_exps.txt'
        if self.SERVER in CONFIDENTIAL_SERVERS:
            today_remote = (
                'huwang@137.203.141.202:/homes/huwang/Projects/ExpLogs/%s' %
                today_exp)
        else:
            today_remote = (
                'wanghuan@155.33.198.138:/home/wanghuan/Projects/ExpLogs/%s' %
                today_exp)
        local_f = 'wh_exps_%s.tmp' % time.time()
        try:
            script_pull = 'scp %s %s' % (today_remote, local_f)
            os.system(script_pull)
        except:
            pass
        with open(local_f, 'a+') as f:
            f.write(self.ExpNote + '\n')
        script_push = 'scp %s %s' % (local_f, today_remote)
        os.system(script_push)
        os.remove(local_f)

    def save_args(self, args):
        with open(pjoin(self.log_path, 'params.json'), 'w') as f:
            json.dump(args.__dict__, f, indent=4)
        with open(pjoin(self.log_path, 'params.yaml'), 'w') as f:
            yaml.dump(args.__dict__, f, indent=4)

    def netprint(self, net, comment=''):
        with open(pjoin(self.log_path, 'model_arch.txt'), 'w') as f:
            if comment:
                print('%s:' % comment, file=f)
            print('%s\n' % str(net), file=f, flush=True)


def adjust_learning_rate(optimizer, epoch, num_epochs, scheduler_type, lr):
    """Sets the learning rate to the initial LR decayed by 2 every 30
    epochs."""
    if scheduler_type == 1:
        new_lr = lr * (0.5**(epoch // 30))
        for param_group in optimizer.param_groups:
            param_group['lr'] = new_lr
    else:
        if epoch in [num_epochs * 0.5, num_epochs * 0.75]:
            for param_group in optimizer.param_groups:
                param_group['lr'] *= 0.1


# +
def strlist_to_list(sstr, ttype):
    """
    example:
    # self.args.stage_pr = [0, 0.3, 0.3, 0.3, 0, ]
    # self.args.skip_layers = ['1.0', '2.0', '2.3', '3.0', '3.5', ]
    turn these into a list of <ttype> (float or str or int etc.)
    """
    if not sstr:
        return sstr
    out = []
    sstr = sstr.strip()
    if sstr.startswith('[') and sstr.endswith(']'):
        sstr = sstr[1:-1]
    for x in sstr.split(','):
        x = x.strip()
        if x:
            x = ttype(x)
            out.append(x)
    return out


def strdict_to_dict(sstr, ttype):
    """'{"1": 0.04, "2": 0.04, "4": 0.03, "5": 0.02, "7": 0.03, }'."""
    if not sstr:
        return sstr
    out = {}
    sstr = sstr.strip()
    if sstr.startswith('{') and sstr.endswith('}'):
        sstr = sstr[1:-1]
    for x in sstr.split(','):
        x = x.strip()
        if x:
            k = x.split(':')[0]
            v = ttype(x.split(':')[1].strip())
            out[k] = v
    return out


def check_path(x):
    if x:
        complete_path = glob.glob(x)
        assert len(complete_path) == 1
        x = complete_path[0]
    return x


def merge_args(args, params_json):
    import json, yaml
    """<args> is from argparser.

    <params_json> is a json/yaml file. merge them, if there is collision, the
    param in <params_json> has a higher priority.
    """
    with open(params_json) as f:
        if params_json.endswith('.json'):
            params = json.load(f)
        elif params_json.endswith('.yaml'):
            params = yaml.load(f, Loader=yaml.FullLoader)
        else:
            raise NotImplementedError
    for k, v in params.items():
        args.__dict__[k] = v
    return args


# -


class PresetLRScheduler(object):
    """Using a manually designed learning rate schedule rules."""

    def __init__(self, decay_schedule):
        # decay_schedule is a dictionary
        # which is for specifying iteration -> lr
        self.decay_schedule = {}
        for (
                k,
                v,
        ) in (decay_schedule.items()
              ):  # a dict, example: {"0":0.001, "30":0.00001, "45":0.000001}
            self.decay_schedule[int(k)] = v
        # print('Using a preset learning rate schedule:')
        # print(self.decay_schedule)

    def __call__(self, optimizer, e):
        epochs = list(self.decay_schedule.keys())
        epochs = sorted(epochs)  # example: [0, 30, 45]
        lr = self.decay_schedule[epochs[-1]]
        for i in range(len(epochs) - 1):
            if epochs[i] <= e < epochs[i + 1]:
                lr = self.decay_schedule[epochs[i]]
                break
        for param_group in optimizer.param_groups:
            param_group['lr'] = lr
        return lr



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from collections import abc
import torch
from torch._six import string_classes
from torch.nn.parallel.scatter_gather import gather
from torch.nn import DataParallel, Sequential, ModuleList
from torch.nn.parallel import DistributedDataParallel
from trailmet.algorithms.algorithms import BaseAlgorithm

__all__ = ['Distillation']


class Distillation(BaseAlgorithm):

    def __init__(self, **kwargs):
        super(Distillation, self).__init__(**kwargs)

        pass


# https://github.com/yoshitomo-matsubara/torchdistill/blob/8be107e00d6a481711e513f45824ecfaa8249c77/torchdistill/common/module_util.py


def check_if_wrapped(model):
    return isinstance(model, (DataParallel, DistributedDataParallel))


def get_module(root_module, module_path):
    module_names = module_path.split('.')
    module = root_module
    for module_name in module_names:
        if not hasattr(module, module_name):
            if isinstance(module, (DataParallel, DistributedDataParallel)):
                module = module.module
                if not hasattr(module, module_name):
                    if (isinstance(module, Sequential)
                            and module_name.lstrip('-').isnumeric()):
                        module = module[int(module_name)]
                    else:
                        raise Exception(
                            '`{}` of `{}` could not be reached in `{}`'.format(
                                module_name, module_path,
                                type(root_module).__name__))
                        # logger.info('`{}` of `{}` could not be reached in `{}`'.format(module_name, module_path,
                        #                                                                type(root_module).__name__))
                else:
                    module = getattr(module, module_name)
            elif (isinstance(module, (Sequential, ModuleList))
                  and module_name.lstrip('-').isnumeric()):
                module = module[int(module_name)]
            else:
                raise Exception(
                    '`{}` of `{}` could not be reached in `{}`'.format(
                        module_name, module_path,
                        type(root_module).__name__))
                # logger.info('`{}` of `{}` could not be reached in `{}`'.format(module_name, module_path,
                #                                                                type(root_module).__name__))
                return None
        else:
            module = getattr(module, module_name)
    return module


# https://github.com/yoshitomo-matsubara/torchdistill/blob/main/torchdistill/core/forward_hook.py


def get_device_index(data):
    if isinstance(data, torch.Tensor):
        device = data.device
        return 'cpu' if device.type == 'cpu' else device.index
    elif isinstance(data, abc.Mapping):
        for key, data in data.items():
            result = get_device_index(data)
            if result is not None:
                return result
    elif isinstance(data, tuple):
        for d in data:
            result = get_device_index(d)
            if result is not None:
                return result
    elif isinstance(data,
                    abc.Sequence) and not isinstance(data, string_classes):
        for d in data:
            result = get_device_index(d)
            if result is not None:
                return result
    return None


def register_forward_hook_with_dict(module, module_path, requires_input,
                                    requires_output, io_dict):
    io_dict[module_path] = dict()

    def forward_hook4input(self, func_input, func_output):
        if isinstance(func_input, tuple) and len(func_input) == 1:
            func_input = func_input[0]

        device_index = get_device_index(func_output)
        sub_io_dict = io_dict[module_path]
        if 'input' not in sub_io_dict:
            sub_io_dict['input'] = dict()
        sub_io_dict['input'][device_index] = func_input

    def forward_hook4output(self, func_input, func_output):
        if isinstance(func_output, tuple) and len(func_output) == 1:
            func_output = func_output[0]

        device_index = get_device_index(func_output)
        sub_io_dict = io_dict[module_path]
        if 'output' not in sub_io_dict:
            sub_io_dict['output'] = dict()
        sub_io_dict['output'][device_index] = func_output

    def forward_hook4io(self, func_input, func_output):
        if isinstance(func_input, tuple) and len(func_input) == 1:
            func_input = func_input[0]
        if isinstance(func_output, tuple) and len(func_output) == 1:
            func_output = func_output[0]

        device_index = get_device_index(func_output)
        sub_io_dict = io_dict[module_path]
        if 'input' not in sub_io_dict:
            sub_io_dict['input'] = dict()

        if 'output' not in sub_io_dict:
            sub_io_dict['output'] = dict()

        sub_io_dict['input'][device_index] = func_input
        sub_io_dict['output'][device_index] = func_output

    if requires_input and not requires_output:
        return module.register_forward_hook(forward_hook4input)
    elif not requires_input and requires_output:
        return module.register_forward_hook(forward_hook4output)
    elif requires_input and requires_output:
        return module.register_forward_hook(forward_hook4io)
    raise ValueError('Either requires_input or requires_output should be True')


class ForwardHookManager(object):
    """
    Example::
        >>> import torch
        >>> from torchvision import models
        >>> from torchdistill.core.forward_hook import ForwardHookManager
        >>> device = torch.device('cpu')
        >>> forward_hook_manager = ForwardHookManager(device)
        >>> model = models.resnet18()
        >>> forward_hook_manager.add_hook(model, 'layer2')
        >>> x = torch.rand(16, 3, 224, 224)
        >>> y = model(x)
        >>> io_dict = forward_hook_manager.pop_io_dict()
        >>> layer2_input_tensor = io_dict['layer2']['input']
        >>> layer2_output_tensor = io_dict['layer2']['output']
    """

    def __init__(self, target_device):
        self.target_device = (torch.device(target_device) if isinstance(
            target_device, str) else target_device)
        self.uses_cuda = self.target_device.type == 'cuda'
        self.io_dict = dict()
        self.hook_list = list()

    def add_hook(self,
                 module,
                 module_path,
                 requires_input=True,
                 requires_output=True):
        unwrapped_module = module.module if check_if_wrapped(
            module) else module
        sub_module = get_module(unwrapped_module, module_path)
        handle = register_forward_hook_with_dict(sub_module, module_path,
                                                 requires_input,
                                                 requires_output, self.io_dict)
        self.hook_list.append((module_path, handle))

    def pop_io_dict(self):
        gathered_io_dict = dict()
        for module_path, module_io_dict in self.io_dict.items():
            gathered_io_dict[module_path] = dict()
            for io_type in list(module_io_dict.keys()):
                sub_dict = module_io_dict.pop(io_type)
                values = [sub_dict[key] for key in sorted(sub_dict.keys())]
                gathered_obj = (gather(values, self.target_device)
                                if self.uses_cuda and len(values) > 1 else
                                values[-1])
                gathered_io_dict[module_path][io_type] = gathered_obj
        return gathered_io_dict

    def pop_io_dict_from_device(self, device):
        device_io_dict = dict()
        device_key = device.index if device.type == 'cuda' else device.type
        for module_path, module_io_dict in self.io_dict.items():
            device_io_dict[module_path] = dict()
            for io_type in list(module_io_dict.keys()):
                sub_dict = module_io_dict[io_type]
                device_io_dict[module_path][io_type] = sub_dict.pop(device_key)
        return device_io_dict

    def change_target_device(self, target_device):
        if self.target_device.type != target_device.type:
            for sub_dict in self.io_dict.values():
                sub_dict.clear()
        self.target_device = target_device

    def clear(self):
        self.io_dict.clear()
        for _, handle in self.hook_list:
            handle.remove()
        self.hook_list.clear()



#!/usr/bin/env python

# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# importing the required packages
import os
from .cifar import CIFAR10Dataset, CIFAR100Dataset
from .imagenet import ImageNetDataset
from .chest import ChestDataset
from .tcga import TCGADataset


class DatasetFactory(object):
    """This class forms the generic wrapper for the different dataset classes.

    The module includes utilities to load datasets, including methods to load
    and fetch popular reference datasets.
    """

    @staticmethod
    def create_dataset(**kwargs):
        """
        Args:
            name(string): dataset name 'CIFAR10', 'CIFAR100', 'ImageNet', 'CHEST',
            root(string):  Root directory of dataset where directory
                   cifar-10-batches-py exists or will be saved
                   to if download is set to True.
        Return:
            dataset(tuple): dataset
        """
        assert 'name' in kwargs, 'should provide dataset name'
        name = kwargs['name']
        assert 'root' in kwargs, 'should provide dataset root'
        if 'CIFAR10' == name:
            obj_dfactory = CIFAR10Dataset(**kwargs)
        elif 'CIFAR100' == name:
            obj_dfactory = CIFAR100Dataset(**kwargs)
        elif 'ImageNet' == name:
            obj_dfactory = ImageNetDataset(**kwargs)
        elif 'CHEST' == name:
            obj_dfactory = ChestDataset(**kwargs)
        elif 'TCGA' == name:
            obj_dfactory = TCGADataset(**kwargs)
        else:
            raise Exception(f"unknown dataset{kwargs['name']}")
        dataset = obj_dfactory.stack_dataset()
        dataset = obj_dfactory.build_dict_info()

        return dataset



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from .chip import Chip
from .chipnet import ChipNet
from .growth_regularisation import Growth_Regularisation
from .network_slimming import Network_Slimming
from .pns import ChannelRounding, Conv2dWrapper, LinearWrapper, BN2dWrapper, SlimPruner
from .prune import BasePruning
from .pruner import pruner_dict
from .utils import (
    update_bn_grad,
    summary_model,
    is_depthwise_conv2d,
    prune_conv2d,
    prune_bn2d,
    prune_fc,
    cal_threshold_by_bn2d_weights,
    mask2idxes,
    top_k_idxes,
    ceil,
    round_up_to_power_of_2,
)



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from .base_model import BaseModel
import torch
import torch.nn as nn
import torch.nn.functional as F


class BasicBlock(nn.Module):

    def __init__(self, in_planes, out_planes, stride, dropRate=0.0):
        super(BasicBlock, self).__init__()
        self.bn1 = nn.BatchNorm2d(in_planes)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv2d(in_planes,
                               out_planes,
                               kernel_size=3,
                               stride=stride,
                               padding=1,
                               bias=False)
        self.bn2 = nn.BatchNorm2d(out_planes)
        self.relu2 = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(out_planes,
                               out_planes,
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               bias=False)
        self.droprate = dropRate
        self.equalInOut = in_planes == out_planes
        self.convShortcut = ((not self.equalInOut) and nn.Conv2d(
            in_planes,
            out_planes,
            kernel_size=1,
            stride=stride,
            padding=0,
            bias=False,
        ) or None)

    def forward(self, x):
        if not self.equalInOut:
            x = self.relu1(self.bn1(x))
        else:
            out = self.relu1(self.bn1(x))
        out = self.relu2(self.bn2(self.conv1(out if self.equalInOut else x)))
        if self.droprate > 0:
            out = F.dropout(out, p=self.droprate, training=self.training)
        out = self.conv2(out)
        return torch.add(x if self.equalInOut else self.convShortcut(x), out)


class NetworkBlock(nn.Module):

    def __init__(self,
                 nb_layers,
                 in_planes,
                 out_planes,
                 block,
                 stride,
                 dropRate=0.0):
        super(NetworkBlock, self).__init__()
        self.layer = self._make_layer(block, in_planes, out_planes, nb_layers,
                                      stride, dropRate)

    def _make_layer(self, block, in_planes, out_planes, nb_layers, stride,
                    dropRate):
        layers = []
        for i in range(int(nb_layers)):
            layers.append(
                block(
                    i == 0 and in_planes or out_planes,
                    out_planes,
                    i == 0 and stride or 1,
                    dropRate,
                ))
        return nn.Sequential(*layers)

    def forward(self, x):
        return self.layer(x)


class WideResNet(BaseModel):

    def __init__(self, depth, num_classes, widen_factor=1, dropRate=0.0):
        super(WideResNet, self).__init__()
        nChannels = [
            16, 16 * widen_factor, 32 * widen_factor, 64 * widen_factor
        ]
        assert (depth - 4) % 6 == 0
        n = (depth - 4) / 6
        block = BasicBlock
        # 1st conv before any network block
        self.conv1 = nn.Conv2d(3,
                               nChannels[0],
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               bias=False)
        # 1st block
        self.block1 = NetworkBlock(n, nChannels[0], nChannels[1], block, 1,
                                   dropRate)
        # 2nd block
        self.block2 = NetworkBlock(n, nChannels[1], nChannels[2], block, 2,
                                   dropRate)
        # 3rd block
        self.block3 = NetworkBlock(n, nChannels[2], nChannels[3], block, 2,
                                   dropRate)
        # global average pooling and classifier
        self.bn1 = nn.BatchNorm2d(nChannels[3])
        self.relu = nn.ReLU(inplace=True)
        self.fc = nn.Linear(nChannels[3], num_classes)
        self.nChannels = nChannels[3]

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight,
                                        mode='fan_out',
                                        nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                m.bias.data.zero_()

    def forward(self, x):
        out = self.conv1(x)
        out1 = self.block1(out)
        out2 = self.block2(out1)
        out3 = self.block3(out2)
        out = self.relu(self.bn1(out3))
        out = F.avg_pool2d(out, 8)
        out = out.view(-1, self.nChannels)

        return self.fc(out)


def make_wideresnet(depth, widen_factor, num_classes, dropRate=0.0):
    """Creates and returns the requested model.

    Parameters
    ----------
    depth : int
    widen_factor : int
    num_classes : int
    dropRate : float, optional

    Returns
    -------
    WideResNet model
    """
    model = WideResNet(depth, num_classes, widen_factor, dropRate)
    return model



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# importing the required packages
from .dataset import BaseDataset
from PIL import Image
import os
import pandas as pd
import numpy as np
from tqdm import tqdm
import torch
from torch.utils.data import Dataset
import gdown
import zipfile
from glob import glob


class TCGA(Dataset):
    """
    `TCGA <https://www.cancer.gov/ccg/research/structural-genomics/tcga/studied-cancers/lung-squamous-cell-carcinoma-study>`_ Dataset.

    References
    ----------

    Parameters
    ----------
        name (string): dataset name 'TCGA', default=None.
        root (string): Root directory where ``splits, images`` exists or will be saved if download flag is set to
        True (default is None).
        mode (string): Mode to run the dataset, options are 'train', 'val', 'test', default='train'.
        train (bool, optional): If True, creates dataset from training set, otherwise
            creates from test set, default=None.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``. Default=None.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it, default=None.
        kfold(int): The fold which you want to use. Possible values are 0-9
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again, default=True.
        split_types (list): the possible values of this parameter are "train", "test" and "val".
            If the split_type contains "val", then shuffle has to be True, default value is None.
        val_fraction (float): If float, should be between 0.0 and 1.0 and represent
        the proportion of the dataset to include in the val split.
        shuffle (bool): Whether or not to shuffle the data before splitting into val from train,
            default is True. If shuffle is true, there should be 'val' in split_types.
        random_seed (int): RandomState instance, default=None.
    """

    def __init__(
        self,
        root=None,
        transform=None,
        target_transform=None,
        kfold=0,
        mode='train',
        download=True,
    ):
        self.labels_dict = {}
        self._kfold = kfold
        self.transform = transform
        self.target_transform = target_transform
        self.class_dict = {'TCGA-LUAD': 0, 'TCGA-LUSC': 1}
        self.split_file = os.path.join(root,
                                       f'tcga_lung/splits_{self._kfold}.csv')
        self.data_path = os.path.join(root, f'512_imgs/fold-{self._kfold}/')
        df = pd.read_csv(self.split_file)

        files = glob(os.path.join(root,
                                  f'512_imgs/fold-{self._kfold}/*/*.png'))

        for filename in files:
            self.labels_dict[filename.split('/')[-1][:-4]] = filename.split(
                '/')[-2]

        self.Train_List, self.Val_List, self.Test_List = (
            df.train.to_list(),
            df.val.dropna().to_list(),
            df.test.dropna().to_list(),
        )

        if mode == 'train':
            self._mode = 'train'
            print('Total Training Sample {}'.format(len(self.Train_List)))
        elif mode == 'val':
            self._mode = 'val'
            print('Total Validating Sample {}'.format(len(self.Val_List)))
        elif mode == 'test':
            self._mode = 'test'
            print('Total test Sample {}'.format(len(self.Test_List)))

    def __len__(self):
        if self._mode == 'train':
            return len(self.Train_List)
        elif self._mode == 'val':
            return len(self.Val_List)
        elif self._mode == 'test':
            return len(self.Test_List)

    def __getitem__(self, idx):
        if self._mode == 'train':
            path = self.Train_List[idx]
            image_path = os.path.join(self.data_path, self.labels_dict[path],
                                      path + '.png')
            img = Image.open(image_path).convert('RGB')
            img = self.transform(img)
            class_name = self.labels_dict[path]
            return img, torch.tensor([self.class_dict[class_name]])
        elif self._mode == 'val':
            path = self.Val_List[idx]
            image_path = os.path.join(self.data_path, self.labels_dict[path],
                                      path + '.png')
            img = Image.open(image_path).convert('RGB')
            img = self.target_transform(img)
            class_name = self.labels_dict[path]
            return img, torch.tensor([self.class_dict[class_name]])
        else:
            path = self.Test_List[idx]
            image_path = os.path.join(self.data_path, self.labels_dict[path],
                                      path + '.png')
            img = Image.open(image_path).convert('RGB')
            img = self.target_transform(img)
            class_name = self.labels_dict[path]
            return img, torch.tensor([self.class_dict[class_name]])


class TCGADataset(BaseDataset):

    def __init__(
        self,
        name=None,
        root=None,
        transform=None,
        target_transform=None,
        kfold=0,
        download=True,
        split_types=None,
        val_fraction=0.2,
        shuffle=True,
        random_seed=None,
    ):
        super(TCGADataset, self).__init__(
            name=name,
            root=root,
            transform=transform,
            target_transform=target_transform,
            download=download,
            split_types=split_types,
            val_fraction=val_fraction,
            shuffle=shuffle,
            random_seed=random_seed,
        )

        self.val_exists = True
        os.makedirs(root, exist_ok=True)
        final_path = os.path.join(root, 'tcga_512')

        if not os.path.exists(final_path + f'/512_imgs/fold-{kfold}'):
            print(
                f'TCGA dataset is not present in {root}. Downloading the dataset'
            )
            gdown.download(
                id='1bnfg9mq-5NwnKjS7ZlVAySooLCTGAjqb',
                output=f'{root}/tcga_512.zip',
                quiet=False,
            )
            os.makedirs(final_path, exist_ok=True)

            # unzipping the dataset
            with zipfile.ZipFile(f'{root}/tcga_512.zip', 'r') as zip_ref:
                for member in tqdm(zip_ref.infolist(), desc='Extracting '):
                    try:
                        zip_ref.extract(member, final_path)
                    except zipfile.error as e:
                        raise Exception(
                            f'Unable to extract the dataset. Please check the path {root}'
                        )
            print('Removing the zip file')
            os.remove(f'{root}/tcga_512.zip')
            print('Done! downloading the dataset.')

            if not os.path.exists(final_path + f'/512_imgs/fold-{kfold}'):
                raise Exception(
                    f'Unable to download the dataset. Please check the path {root}'
                )
        else:
            print(f'TCGA dataset is present in {final_path}')

        if not os.path.exists(
                os.path.join(root, 'tcga_512', 'tcga_lung',
                             f'splits_{kfold}.csv')):
            print(
                f'TCGA split file is not present in {root}. Downloading the split file'
            )
            gdown.download(
                id='1xBxLz2iToaHaJJovml7Bf4NcJRS_x2CQ',
                output=f'{root}/tcga_512/tcga_lung.zip',
                quiet=False,
            )
            os.makedirs(os.path.join(root, 'tcga_512', 'tcga_lung'),
                        exist_ok=True)

            # unzipping the dataset
            with zipfile.ZipFile(f'{root}/tcga_512/tcga_lung.zip',
                                 'r') as zip_ref:
                for member in tqdm(zip_ref.infolist(), desc='Extracting '):
                    try:
                        zip_ref.extract(member, final_path)
                    except zipfile.error as e:
                        raise Exception(
                            f'Unable to extract the split files. Please check the path {root}'
                        )
            print('Removing the zip file')
            os.remove(f'{root}/tcga_512/tcga_lung.zip')
            print('Done! downloading the split files.')

            if not os.path.exists(
                    os.path.join(root, 'tcga_512', 'tcga_lung',
                                 f'splits_{kfold}.csv')):
                raise Exception(
                    f'Unable to download the split files. Please check the path {root}'
                )
        else:
            print(f'TCGA split files is present in {final_path}')

        dataset = TCGA
        self.dataset_dict = {}

        for item in self.split_types:
            dataset_type = item
            if item == 'val' and not self.val_exists:
                self.dataset_dict[dataset_type] = None
                continue
            data = dataset(
                root=final_path,
                mode=item,
                kfold=kfold,
                transform=self.transform[item],
                target_transform=self.target_transform[item],
            )
            self.dataset_dict[dataset_type] = data

    def build_dict_info(self):
        """
        Behavior:
            This function creates info key in the output dictionary. The info key contains details related to the size
            of the training, validation and test datasets. Further, it can be used to define any additional information
            necessary for the user.
        Returns:
            dataset_dict (dict): Updated with info key that contains details related to the data splits
        """
        self.dataset_dict['info'] = {}
        self.dataset_dict['info']['train_size'] = len(
            self.dataset_dict['train'])
        self.dataset_dict['info']['val_size'] = len(self.dataset_dict['val'])
        self.dataset_dict['info']['test_size'] = len(self.dataset_dict['test'])
        self.dataset_dict['info']['note'] = ''
        return self.dataset_dict



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import torch.nn as nn
import numpy as np
import scipy.optimize as optim
import warnings
from trailmet.algorithms.quantize.quantize import BaseQuantization, RoundSTE
from trailmet.utils import lp_loss

__all__ = [
    'UniformAffineQuantizer',
    'AdaRoundQuantizer',
    'BitSplitQuantizer',
    'ActQuantizer',
    'QuantizationBase',
    'UniformQuantization',
    'ClippedUniformQuantization',
    'FixedClipValueQuantization',
    'MaxAbsStaticQuantization',
    'LearnedStepSizeQuantization',
    'LpNormQuantization',
]
"""Quantization classes:-

[BRECQ]
    - UniformAffineQuantizer
    - AdaRoundQuantizer
[BitSplit]
    - BitSplitQuantizer
    - ActQuantizer
[LAPQ]
    - QuantizationBase
        - UniformQuantization
            - ClippedUniformQuantization
                - FixedClipValueQuantization
                - MaxAbsStaticQuantization
                - LearnedStepSizeQuantization
                - LpNormQuantization
"""


class UniformAffineQuantizer(nn.Module):
    """PyTorch Function that can be used for asymmetric quantization (uniform
    affine quantization).

    Quantizes its argument in the forward pass, passes the gradient 'straight
    through' on the backward pass, ignoring the quantization that occurred.
    Based on
    https://arxiv.org/abs/1806.08342.
    :param n_bits: number of bit for quantization
    :param symmetric: if True, the zero_point should always be 0
    :param channel_wise: if True, compute scale and zero_point in each channel
    :param scale_method: determines the quantization scale and zero point
    """

    def __init__(
        self,
        n_bits: int = 8,
        symmetric: bool = False,
        channel_wise: bool = False,
        scale_method: str = 'max',
        leaf_param: bool = False,
    ):
        super(UniformAffineQuantizer, self).__init__()
        self.sym = symmetric
        assert 2 <= n_bits <= 8, 'bitwidth not supported'
        self.n_bits = n_bits
        self.n_levels = 2**self.n_bits
        self.delta = None
        self.zero_point = None
        self.inited = False
        self.leaf_param = leaf_param
        self.channel_wise = channel_wise
        self.scale_method = scale_method

    def forward(self, x: torch.Tensor):
        if self.inited is False:
            if self.leaf_param:
                delta, self.zero_point = self.init_quantization_scale(
                    x, self.channel_wise)
                self.delta = torch.nn.Parameter(delta)
                # self.zero_point = torch.nn.Parameter(self.zero_point)
            else:
                self.delta, self.zero_point = self.init_quantization_scale(
                    x, self.channel_wise)
            self.inited = True

        # start quantization
        # x_int = BQ.round_ste(x / self.delta) + self.zero_point
        x_int = RoundSTE.apply(x / self.delta) + self.zero_point
        x_quant = torch.clamp(x_int, 0, self.n_levels - 1)
        x_dequant = (x_quant - self.zero_point) * self.delta
        return x_dequant

    def init_quantization_scale(self,
                                x: torch.Tensor,
                                channel_wise: bool = False):
        delta, zero_point = None, None
        if channel_wise:
            x_clone = x.clone().detach()
            n_channels = x_clone.shape[0]
            if len(x.shape) == 4:
                x_max = x_clone.abs().max(dim=-1)[0].max(dim=-1)[0].max(
                    dim=-1)[0]
            else:
                x_max = x_clone.abs().max(dim=-1)[0]
            delta = x_max.clone()
            zero_point = x_max.clone()
            # determine the scale and zero point channel-by-channel
            for c in range(n_channels):
                delta[c], zero_point[c] = self.init_quantization_scale(
                    x_clone[c], channel_wise=False)
            if len(x.shape) == 4:
                delta = delta.view(-1, 1, 1, 1)
                zero_point = zero_point.view(-1, 1, 1, 1)
            else:
                delta = delta.view(-1, 1)
                zero_point = zero_point.view(-1, 1)
        else:
            if 'max' in self.scale_method:
                x_min = min(x.min().item(), 0)
                x_max = max(x.max().item(), 0)
                if 'scale' in self.scale_method:
                    x_min = x_min * (self.n_bits + 2) / 8
                    x_max = x_max * (self.n_bits + 2) / 8

                x_absmax = max(abs(x_min), x_max)
                if self.sym:
                    x_min, x_max = -x_absmax if x_min < 0 else 0, x_absmax

                delta = float(x_max - x_min) / (self.n_levels - 1)
                if delta < 1e-8:
                    warnings.warn(
                        'Quantization range close to zero: [{}, {}]'.format(
                            x_min, x_max))
                    delta = 1e-8

                zero_point = torch.round(-x_min / delta)
                delta = torch.tensor(delta).type_as(x)

            elif self.scale_method == 'mse':
                # For Lp norm minimization as described in LAPQ
                # https://arxiv.org/abs/1911.07190
                x_max = x.max()
                x_min = x.min()
                best_score = 1e10
                for i in range(80):
                    new_max = x_max * (1.0 - (i * 0.01))
                    new_min = x_min * (1.0 - (i * 0.01))
                    x_q = self.quantize(x, new_max, new_min)
                    score = lp_loss(pred=x, tgt=x_q, p=2.4, reduction='all')
                    if score < best_score:
                        best_score = score
                        delta = (new_max - new_min) / (2**self.n_bits - 1)
                        zero_point = torch.round(-new_min / delta)
            else:
                raise NotImplementedError

        return delta, zero_point

    def quantize(self, x, max, min):
        delta = (max - min) / (2**self.n_bits - 1)
        zero_point = torch.round(-min / delta)
        # we assume weight quantization is always signed
        x_int = torch.round(x / delta)
        x_quant = torch.clamp(x_int + zero_point, 0, self.n_levels - 1)
        x_float_q = (x_quant - zero_point) * delta
        return x_float_q

    def bitwidth_refactor(self, refactored_bit: int):
        assert 2 <= refactored_bit <= 8, 'bitwidth not supported'
        self.n_bits = refactored_bit
        self.n_levels = 2**self.n_bits

    def extra_repr(self):
        s = (
            'bit={n_bits}, scale_method={scale_method}, symmetric={sym}, channel_wise={channel_wise},'
            ' leaf_param={leaf_param}')
        return s.format(**self.__dict__)


class AdaRoundQuantizer(nn.Module):
    """Adaptive Rounding Quantizer, used to optimize the rounding policy by
    reconstructing the intermediate output.

    Based on Up or Down? Adaptive Rounding for Post-Training Quantization:
    https://arxiv.org/abs/2004.10568
    : https: //arxiv.org/abs/2004.10568
    :param uaq: UniformAffineQuantizer, used to initialize quantization
        parameters in this quantizer
    :param round_mode: controls the forward pass in this quantizer
    :param weight_tensor: initialize alpha
    """

    def __init__(
        self,
        uaq: UniformAffineQuantizer,
        weight_tensor: torch.Tensor,
        round_mode='learned_round_sigmoid',
    ):
        super(AdaRoundQuantizer, self).__init__()
        # copying all attributes from UniformAffineQuantizer
        self.n_bits = uaq.n_bits
        self.sym = uaq.sym
        self.delta = uaq.delta
        self.zero_point = uaq.zero_point
        self.n_levels = uaq.n_levels

        self.round_mode = round_mode
        self.alpha = None
        self.soft_targets = False

        # params for sigmoid function
        self.gamma, self.zeta = -0.1, 1.1
        self.beta = 2 / 3
        self.init_alpha(x=weight_tensor.clone())

    def forward(self, x):
        if self.round_mode == 'nearest':
            x_int = torch.round(x / self.delta)
        elif self.round_mode == 'nearest_ste':
            # x_int = BQ.round_ste(x / self.delta)
            x_int = RoundSTE.apply(x / self.delta)
        elif self.round_mode == 'stochastic':
            x_floor = torch.floor(x / self.delta)
            rest = (x / self.delta) - x_floor  # rest of rounding
            x_int = x_floor + torch.bernoulli(rest)
            print('Draw stochastic sample')
        elif self.round_mode == 'learned_hard_sigmoid':
            x_floor = torch.floor(x / self.delta)
            if self.soft_targets:
                x_int = x_floor + self.get_soft_targets()
            else:
                x_int = x_floor + (self.alpha >= 0).float()
        else:
            raise ValueError('Wrong rounding mode')

        x_quant = torch.clamp(x_int + self.zero_point, 0, self.n_levels - 1)
        x_float_q = (x_quant - self.zero_point) * self.delta

        return x_float_q

    def get_soft_targets(self):
        return torch.clamp(
            torch.sigmoid(self.alpha) * (self.zeta - self.gamma) + self.gamma,
            0, 1)

    def init_alpha(self, x: torch.Tensor):
        x_floor = torch.floor(x / self.delta)
        if self.round_mode == 'learned_hard_sigmoid':
            # print('Init alpha to be FP32')
            rest = (x / self.delta) - x_floor  # rest of rounding [0, 1)
            alpha = -torch.log(
                (self.zeta - self.gamma) /
                (rest - self.gamma) - 1)  # => sigmoid(alpha) = rest
            self.alpha = nn.Parameter(alpha)
        else:
            raise NotImplementedError


class BitSplitQuantizer(object):

    def __init__(self, W: np.ndarray, bitwidth):
        self.W = W
        self.bitwidth = bitwidth

    @staticmethod
    def splitWeightInteger(Q, bitwidth):
        """Split low-bit weight integers into a list of ternary weights."""
        Q_sign = np.sign(Q)
        Q_abs = np.abs(Q)
        B_sav = []
        for idx in range(bitwidth - 1):
            B = Q_abs - Q_abs.astype(np.int) // 2 * 2  # get current last bit
            B *= Q_sign
            B_sav.append(B)
            Q_abs = (Q_abs.astype(np.int) // 2).astype(
                np.float32)  # Q_abs >> 1
        return B_sav[::-1]

    @staticmethod
    def splitWeightVector(W, alpha):
        """Get the optimal ternary vector, given the quantization scale
        alpha."""
        B = W.copy()
        B = (B >= 0).astype(np.float32) * 2 - 1
        abs_W2 = np.abs(W) * 2
        B[abs_W2 < alpha[:, np.newaxis]] = 0
        return B

    @staticmethod
    def stitch(B_sav):
        """Stitch a list of ternary vectors into a integer vectors."""
        B_sum = B_sav[0].copy()
        for idx in range(1, len(B_sav)):
            B_sum += B_sav[idx] / (2**idx)
        return B_sum

    @staticmethod
    def stitchExclusive(B_sav, bit):
        """Stitch a list of ternary vectors into a integer vectors.

        The i-th bit is excluded.
        """
        # NOTE: the position of the decimal point is not at the end. E.g. 4-bit fixed-point numbers could be something like:
        # +1.01, -0.11, +1.10, ... instead of +101, -011, +110, ... .
        mask = [1.0] * len(B_sav)
        mask[bit] = 0
        B_sum = B_sav[0] * mask[0]
        for idx in range(1, len(B_sav)):
            B_sum += B_sav[idx] * mask[idx] / (2**idx)
        return B_sum

    def fwa(self):
        """Fixed-point Weight Approximation.

        Minimize the MSE quantization error to find the optimal quantization
        scales Given quantization scale, round-off is used for fixed-point
        quantization
        """
        max_val = 2**(self.bitwidth - 1) - 1
        alpha = np.abs(self.W).max(axis=1) / max_val
        alpha_old = alpha * 1.1
        while np.linalg.norm(alpha - alpha_old) > 1e-9:
            q = self.W / alpha[:, np.newaxis]
            q = np.round(q)
            q = np.clip(q, -max_val, max_val)
            alpha_old = alpha
            alpha = np.sum(self.W * q, axis=1) / np.sum(q * q, axis=1)
        return q, alpha

    def ofwa(self, max_epoch=50):
        """Optimal Fixed Point Weight Approximation Method.

        Minimize weight matrix reconstruction error using bit-split strategy.
        Given quantization scale, we find the 'optimal' low-bit weights that
        minimizes the weight quantization error (instead of using round-off).
        Initialized by "fwa".
        """
        assert 2 <= self.bitwidth <= 16
        Q, alpha = self.fwa()
        B_sav = self.splitWeightInteger(Q, self.bitwidth)
        alpha *= 2**(self.bitwidth - 2)
        # NOTE: the position of the decimal point is not at the end.
        # E.g. 4-bit fixed-point numbers could be something like:
        # +1.01, -0.11, +1.10, ... instead of +101, -011, +110, ... .
        ### iterative optimization
        for _ in range(max_epoch):
            # print(_)
            alpha_old = np.copy(alpha)
            B_sum = self.stitch(B_sav)
            # given Ws, optimize alpha
            alpha = np.sum(self.W * B_sum, axis=1) / np.sum(B_sum * B_sum,
                                                            axis=1)
            if np.linalg.norm(alpha_old - alpha) <= 1e-9:
                break
            # given alpha, optimize Ws
            for bit in range(self.bitwidth - 1):
                W_res = self.W - self.stitchExclusive(
                    B_sav, bit) * alpha[:, np.newaxis]
                B = self.splitWeightVector(W_res * (2**bit), alpha)
                B_sav[bit] = B
        B_sum = self.stitch(B_sav)
        return B_sav, B_sum, alpha

    def ofwa_rr(self, X: np.ndarray, Y: np.ndarray, max_epoch=100):
        """Optimal Fixed Point Weight Approximation with Response
        Reconstruction.

        Minimize activation matrix reconstruction error using bit-split strategy.
        Initialized by "ofwa".
        :X: K,C,d,d
        :Y: K,M
        :B: M,N    M kernels
        objective:
        min(Y-XWA)^2
        """
        # X: K,N   (N=C*d*d)
        B_sav, _, alpha = self.ofwa()
        X = X.reshape(X.shape[0], -1)
        K, N = X.shape
        A = np.dot(X.T, X)  # N,N
        for epoch in range(max_epoch):
            # given Bi, optimize alpha
            B_sum = self.stitch(B_sav)
            XB = np.dot(X, B_sum.T)  # k,m
            alpha = np.einsum('ij,ij->j', Y, XB)
            alpha = alpha / np.einsum('ij,ij->j', XB, XB)
            # given alpha, optimize Bi
            for bit in range(self.bitwidth - 1):
                B = B_sav[bit]
                B_others = self.stitchExclusive(B_sav, bit) * alpha[:,
                                                                    np.newaxis]
                Y_res = Y - np.dot(X, B_others.T)
                T = np.dot(Y_res.T, X)  # M,N
                ## fix alpha, optimize B
                # parallel degree: M
                for n in range(N):
                    B[:, n] = 0
                    ABn = np.dot(A[n], B.T)
                    lump = 2 * (ABn * (alpha / (2**bit)) - T[:, n])  # M
                    B[:, n] = -np.sign(lump)
                    B[np.abs(lump) < (alpha / (2**bit)) * A[n, n], n] = 0
        B_sum = self.stitch(B_sav)
        return B_sum, alpha

    def ofwa_rr_dw(self, X: np.ndarray, Y: np.ndarray, max_epoch=100):
        """
        # X: K,M,d,d
        # Y: K,M
        # B: M,N    M kernels
        objective:
        min(Y-XWA)^2
        """
        # X: M,K,9   (N=d*d)
        B_sav, _, alpha = self.ofwa()
        X = np.transpose(X.reshape(X.shape[0], X.shape[1], -1),
                         (1, 0, 2))  # M, K, 9
        As = np.matmul(np.transpose(X, (0, 2, 1)), X)  # M, 9, 9

        alpha_bk = alpha
        for epoch in range(max_epoch):
            # given Bi, optimize alpha
            B_sum = self.stitch(B_sav)
            XB = np.matmul(X, np.expand_dims(B_sum, axis=2))  # M, K, 1
            XB = np.squeeze(XB, axis=2)  # M, K
            XB = XB.T

            alpha = np.einsum('ij,ij->j', Y, XB)
            alpha = alpha / np.einsum('ij,ij->j', XB, XB)
            nan_pos = np.isnan(alpha)
            alpha[nan_pos] = alpha_bk[nan_pos]

            # given alpha, optimize Bi
            for bit in range(self.bitwidth - 1):
                B = B_sav[bit]
                B_others = self.stitchExclusive(B_sav, bit) * alpha[:,
                                                                    np.newaxis]
                Y_res = (Y - np.squeeze(
                    np.matmul(X, np.expand_dims(B_others, axis=2)), axis=2).T
                         )  # Y_res = Y - np.dot(X, B_others.T)

                T = np.squeeze(np.matmul(np.expand_dims(Y_res.T, axis=1), X),
                               axis=1)  # T = np.dot(Y_res.T, X) # M,N
                ## fix alpha, optimize B
                # parallel degree: M
                for n in range(9):  # N=9
                    B[:, n] = 0
                    ABn = np.diagonal(np.dot(
                        As[:, n], B.T))  # M #ABn = np.dot(A[n], B.T)
                    lump = 2 * (ABn * (alpha / (2**bit)) - T[:, n])  # M
                    B[:, n] = -np.sign(lump)

                    B[np.abs(lump) < (alpha / (2**bit)) * As[:, n, n], n] = 0

        B_sum = self.stitch(B_sav)

        return B_sum, alpha


class ActQuantizer(nn.Module):

    def __init__(self, islinear=False, bit_width=8):
        super(ActQuantizer, self).__init__()
        # self.scale = None
        self.in_scale = None
        self.out_scale = None
        self.signed = islinear
        self.bit_width = bit_width
        self.set_bitwidth(self.bit_width)

    def set_bitwidth(self, bit_width):
        self.bit_width = bit_width
        if self.signed:
            self.max_val = (1 << (self.bit_width - 1)) - 1
            self.min_val = -self.max_val
        else:
            self.max_val = (1 << self.bit_width) - 1
            self.min_val = 0

    def set_scale(self, scale):
        self.set_inscale(scale)
        self.set_outscale(scale)

    def set_inscale(self, in_scale):
        self.in_scale = in_scale
        if isinstance(self.in_scale, (float, np.float32, np.float64)):
            pass
        else:
            self.in_scale = torch.tensor(self.in_scale).view(1, -1, 1, 1)

    def set_outscale(self, out_scale):
        self.out_scale = out_scale
        if isinstance(self.out_scale, (float, np.float32, np.float64)):
            pass
        else:
            self.out_scale = torch.tensor(self.out_scale).view(1, -1, 1, 1)

    def init_quantization(self, x):
        assert np.min(x) >= 0
        circle_detection_queue = [
            0,
        ] * 5
        alpha = np.max(np.fabs(x)) / self.max_val
        alpha_old = alpha * 0
        n_iter = 0
        circle_detection_queue[n_iter] = alpha
        while np.sum(alpha != alpha_old):
            q = x / alpha
            q = np.clip(np.round(q), self.min_val, self.max_val)
            alpha_old = alpha
            alpha = np.sum(x * q) / np.sum(q * q)
            if alpha in circle_detection_queue:
                break
            n_iter += 1
            circle_detection_queue[n_iter % 5] = alpha
        return alpha

    def forward(self, x):
        if self.in_scale is None:
            assert self.out_scale is None
            return x
        if not isinstance(self.in_scale, (float, np.float32, np.float64)):
            self.in_scale = self.in_scale.to(x.device)
        if not isinstance(self.out_scale, (float, np.float32, np.float64)):
            self.out_scale = self.out_scale.to(x.device)
        # return torch.clamp(torch.round(x/self.in_scale), self.min_val, self.max_val) * self.out_scale
        return (torch.clamp(RoundSTE.apply(x / self.in_scale), self.min_val,
                            self.max_val) * self.out_scale)


class QuantizationBase(object):

    def __init__(self, module, num_bits):
        self.module = module
        self.num_bits = num_bits
        self.num_bins = int(2**num_bits)
        self.opt_params = {}
        self.named_params = []

    def register_buffer(self, name, value):
        if hasattr(self.module, name):
            delattr(self.module, name)
        self.module.register_buffer(name, value)
        setattr(self, name, getattr(self.module, name))

    def register_parameter(self, name, value):
        if hasattr(self.module, name):
            delattr(self.module, name)
        self.module.register_parameter(name, nn.Parameter(value))
        setattr(self, name, getattr(self.module, name))

        self.named_params.append((name, getattr(self.module, name)))

    def __add_optim_params__(self, optim_type, dataset, params):
        learnable_params = [
            d for n, d in params if n in self.learned_parameters()
        ]
        self.opt_params[optim_type + '_' + dataset] = learnable_params

    def optim_parameters(self):
        return self.opt_params

    def loggable_parameters(self):
        return self.named_parameters()

    def named_parameters(self):
        named_params = [(n, p) for n, p in self.named_params
                        if n in self.learned_parameters()]
        return named_params

    @staticmethod
    def learned_parameters():
        return []


class UniformQuantization(QuantizationBase):

    def __init__(self,
                 module,
                 num_bits,
                 symmetric,
                 uint=False,
                 stochastic=False,
                 tails=False):
        super(UniformQuantization, self).__init__(module, num_bits)
        if not symmetric and not uint:
            raise RuntimeError(
                "Can't perform integer quantization on non symmetric distributions."
            )
        self.symmetric = symmetric
        self.uint = uint
        self.stochastic = stochastic
        self.tails = tails
        if uint:
            self.qmax = 2**self.num_bits - 1
            self.qmin = 0
        else:
            self.qmax = 2**(self.num_bits - 1) - 1
            self.qmin = -self.qmax - 1
        if tails:
            self.qmax -= 0.5 + 1e-6
            self.qmin -= 0.5

    def __quantize__(self, tensor, alpha):
        delta = (2 if self.symmetric else 1) * alpha / (self.num_bins - 1)
        delta = max(delta, 1e-8)
        # quantize
        if self.uint and self.symmetric:
            t_q = (tensor + alpha) / delta
        else:
            t_q = tensor / delta
        # stochastic rounding
        if self.stochastic and self.module.training:
            with torch.no_grad():
                noise = t_q.new_empty(t_q.shape).uniform_(-0.5, 0.5)
                t_q += noise
        # clamp and round
        t_q = torch.clamp(t_q, self.qmin, self.qmax)
        t_q = RoundSTE.apply(t_q)
        assert torch.unique(t_q).shape[0] <= self.num_bins
        # de-quantize
        if self.uint and self.symmetric:
            t_q = t_q * delta - alpha
        else:
            t_q = t_q * delta
        return t_q

    def __quantize_gemmlowp__(self, tensor, min_, max_):
        assert self.uint is True
        delta = (max_ - min_) / (self.num_bins - 1)
        delta = max(delta, 1e-8)
        # quantize
        t_q = (tensor - min_) / delta
        # stochastic rounding
        if self.stochastic and self.module.training:
            with torch.no_grad():
                noise = t_q.new_empty(t_q.shape).uniform_(-0.5, 0.5)
                t_q += noise
        # clamp and round
        t_q = torch.clamp(t_q, self.qmin, self.qmax)
        t_q = RoundSTE.apply(t_q)
        assert torch.unique(t_q).shape[0] <= self.num_bins
        # de-quantize
        t_q = t_q * delta + min_
        return t_q

    def __for_repr__(self):
        return [
            ('bits', self.num_bits),
            ('symmetric', self.symmetric),
            ('tails', self.tails),
        ]

    def __repr__(self):
        s = '{} - ['.format(type(self).__name__)
        for name, value in self.__for_repr__():
            s += '{}: {}, '.format(name, value)
        return s + ']'


class ClippedUniformQuantization(UniformQuantization):
    alpha_param_name = 'alpha'

    def __init__(self,
                 module,
                 num_bits,
                 symmetric,
                 uint=False,
                 stochastic=False,
                 tails=False):
        super(ClippedUniformQuantization,
              self).__init__(module, num_bits, symmetric, uint, stochastic,
                             tails)

    def __call__(self, tensor):
        t_q = self.__quantize__(tensor, self.alpha)
        return t_q

    def __for_repr__(self):
        rpr = super(ClippedUniformQuantization, self).__for_repr__()
        return [(
            self.alpha_param_name,
            '{:.4f}'.format(getattr(self, self.alpha_param_name).item()),
        )] + rpr


class FixedClipValueQuantization(ClippedUniformQuantization):

    def __init__(self,
                 module,
                 num_bits,
                 symmetric,
                 uint=False,
                 stochastic=False,
                 kwargs={}):
        super(FixedClipValueQuantization,
              self).__init__(module, num_bits, symmetric, uint, stochastic)
        self.clip_value = kwargs['clip_value']
        self.device = kwargs['device']
        with torch.no_grad():
            self.register_buffer(
                self.alpha_param_name,
                torch.tensor([self.clip_value],
                             dtype=torch.float32).to(self.device),
            )


class MaxAbsStaticQuantization(ClippedUniformQuantization):

    def __init__(
        self,
        module,
        tensor,
        num_bits,
        symmetric,
        uint=False,
        stochastic=False,
        kwargs={},
    ):
        super(MaxAbsStaticQuantization,
              self).__init__(module, num_bits, symmetric, uint, stochastic)

        with torch.no_grad():
            self.register_buffer(self.alpha_param_name,
                                 tensor.new_tensor([tensor.abs().max()]))


class LearnedStepSizeQuantization(ClippedUniformQuantization):

    def __init__(self,
                 module,
                 tensor,
                 num_bits,
                 symmetric,
                 uint=False,
                 stochastic=False,
                 **kwargs):
        super(LearnedStepSizeQuantization,
              self).__init__(module, num_bits, symmetric, uint, stochastic)

        with torch.no_grad():
            maxabs = tensor.abs().max()

        self.register_parameter(self.alpha_param_name,
                                tensor.new_tensor([maxabs]))

        self.__create_optim_params__()

    def __create_optim_params__(self):
        # TODO: create default configuration
        self.__add_optim_params__(
            'SGD',
            'imagenet',
            [(
                self.alpha_param_name,
                {
                    'params': [getattr(self, self.alpha_param_name)],
                    'lr': 1e-3,
                    'momentum': 0,
                    'weight_decay': 0,
                },
            )],
        )
        self.__add_optim_params__(
            'SGD',
            'cifar10',
            [(
                self.alpha_param_name,
                {
                    'params': [getattr(self, self.alpha_param_name)],
                    'lr': 1e-1,
                    'momentum': 0,
                    'weight_decay': 0,
                },
            )],
        )

    @staticmethod
    def learned_parameters():
        return [LearnedStepSizeQuantization.alpha_param_name]


class LpNormQuantization(ClippedUniformQuantization):

    def __init__(
        self,
        module,
        tensor,
        num_bits,
        symmetric,
        uint=False,
        stochastic=False,
        tails=False,
        kwargs={},
    ):
        super(LpNormQuantization, self).__init__(module, num_bits, symmetric,
                                                 uint, stochastic, tails)

        self.p = kwargs['lp']
        with torch.no_grad():
            opt_alpha = optim.minimize_scalar(
                lambda alpha: self.estimate_quant_error(alpha, tensor),
                bounds=(tensor.min().item(), tensor.max().item()),
            ).x

        self.register_buffer(self.alpha_param_name,
                             tensor.new_tensor([opt_alpha]))

    def estimate_quant_error(self, alpha, x):
        xq = self.__quantize__(x, alpha)
        err = torch.mean(torch.abs(xq - x)**self.p)
        return err.item()



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# importing the required packages
from trailmet.algorithms.prune.prune import BasePruning
import torch
import numpy as np
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt
import pandas as pd

import logging
from datetime import datetime
from tqdm import tqdm
import wandb
import pandas as pd
import numpy as np
import os
import time

from trailmet.utils import (
    AverageMeter,
    accuracy,
    save_checkpoint,
    seed_everything,
    adjust_learning_rate,
)

logger = logging.getLogger(__name__)

seed_everything(43)


class PrunableBatchNorm2d(torch.nn.BatchNorm2d):
    """Wrapper class for batch norm layer to make it prunable."""

    def __init__(self, num_features: int, conv_module: nn.Conv2d = None):
        super(PrunableBatchNorm2d, self).__init__(num_features=num_features)
        self.is_imp = False
        self.is_pruned = False
        self.num_gates = num_features
        self.zeta = nn.Parameter(torch.rand(num_features) * 0.01)
        self.pruned_zeta = torch.ones_like(self.zeta)
        if conv_module is not None:

            def fo_hook(module, in_tensor, out_tensor):
                module.num_input_active_channels = ((in_tensor[0].sum(
                    (0, 2, 3)) > 0).sum().item())
                module.output_area = out_tensor.size(2) * out_tensor.size(3)

            conv_module.register_forward_hook(fo_hook)
        self._conv_module = conv_module
        beta = 1.0
        gamma = 2.0
        for n, x in zip(
            ('beta', 'gamma'),
            (torch.tensor([x], requires_grad=False) for x in (beta, gamma)),
        ):
            self.register_buffer(
                n, x)  # self.beta will be created (same for gamma, zeta)

    def forward(self, input_data):
        out = super(PrunableBatchNorm2d, self).forward(input_data)
        z = self.pruned_zeta if self.is_pruned else self.get_zeta_t()
        out *= z[
            None, :, None,
            None]  # broadcast the mask to all samples in the batch, and all locations
        return out

    def get_zeta_i(self):
        """Returns the zeta_i by applying generalized logistic transformation
        on zeta."""
        return self.__generalized_logistic(self.zeta)

    def get_zeta_t(self):
        """Returns zeta_t by applying continuous heaviside transformation on
        zeta_i."""
        zeta_i = self.get_zeta_i()
        return self.__continuous_heaviside(zeta_i)

    def set_beta_gamma(self, beta, gamma):
        """Sets the values of beta and gamma."""
        self.beta.data.copy_(torch.Tensor([beta]))
        self.gamma.data.copy_(torch.Tensor([gamma]))

    def __generalized_logistic(self, x):
        return 1.0 / (1.0 + torch.exp(-self.beta * x))

    def __continuous_heaviside(self, x):
        return 1 - torch.exp(-self.gamma * x) + x * torch.exp(-self.gamma)

    def prune(self, threshold):
        self.is_pruned = True
        self.pruned_zeta = (self.get_zeta_t() > threshold).float()
        self.zeta.requires_grad = False

    def unprune(self):
        self.is_pruned = False
        self.zeta.requires_grad = True

    def get_params_count(self):
        total_conv_params = (self._conv_module.in_channels *
                             self.pruned_zeta.shape[0] *
                             self._conv_module.kernel_size[0] *
                             self._conv_module.kernel_size[1])
        bn_params = self.num_gates * 2
        active_bn_params = self.pruned_zeta.sum().item() * 2
        active_conv_params = (self._conv_module.num_input_active_channels *
                              self.pruned_zeta.sum().item() *
                              self._conv_module.kernel_size[0] *
                              self._conv_module.kernel_size[1])
        return active_conv_params + active_bn_params, total_conv_params + bn_params

    def get_volume(self):
        total_volume = self._conv_module.output_area * self.num_gates
        active_volume = self._conv_module.output_area * self.pruned_zeta.sum(
        ).item()
        return active_volume, total_volume

    def get_flops(self):
        k_area = self._conv_module.kernel_size[
            0] * self._conv_module.kernel_size[1]
        total_flops = (self._conv_module.output_area * self.num_gates *
                       self._conv_module.in_channels * k_area)
        active_flops = (self._conv_module.output_area *
                        self.pruned_zeta.sum().item() *
                        self._conv_module.num_input_active_channels * k_area)
        return active_flops, total_flops

    @staticmethod
    def from_batchnorm(bn_module, conv_module):
        new_bn = PrunableBatchNorm2d(bn_module.num_features, conv_module)
        return new_bn, conv_module


class ModuleInjection:
    pruning_method = 'full'
    prunable_modules = []

    @staticmethod
    def make_prunable(conv_module, bn_module):
        """Make a (conv, bn) sequence prunable.

        :param conv_module: A Conv2d module
        :param bn_module: The BatchNorm2d module following the Conv2d above
        :param prune_before_bn: Whether the pruning gates will be applied
            before or after the Batch Norm
        :return: a pair (conv, bn) that can be trained to
        """
        if ModuleInjection.pruning_method == 'full':
            return conv_module, bn_module
        new_bn, conv_module = PrunableBatchNorm2d.from_batchnorm(
            bn_module, conv_module=conv_module)
        ModuleInjection.prunable_modules.append(new_bn)
        return conv_module, new_bn


class ChipNet(BasePruning):
    """Class to compress models using chipnet method."""

    def __init__(self, model, dataloaders, **kwargs):
        super(ChipNet, self).__init__(**kwargs)
        self.model = model
        self.dataloaders = dataloaders
        self.kwargs = kwargs
        self.budget_type = self.kwargs['CHIPNET_ARGS'].get(
            'BUDGET_TYPE', 'channel_ratio')
        self.target_budget = self.kwargs['CHIPNET_ARGS'].get(
            'TARGET_BUDGET', 0.5)
        self.steepness = self.kwargs['CHIPNET_ARGS'].get('STEEPNESS', 10)
        self.budget_loss_weightage = self.kwargs['CHIPNET_ARGS'].get(
            'BUDGET_LOSS_WEIGHTAGE', 30)
        self.crispness_loss_weightage = self.kwargs['CHIPNET_ARGS'].get(
            'CRISPNESS_LOSS_WEIGHTAGE', 10)
        self.b_inc = self.kwargs['CHIPNET_ARGS'].get('BETA_INCREMENT', 5.0)
        self.g_inc = self.kwargs['CHIPNET_ARGS'].get('GAMMA_INCREMENT', 2.0)
        self.target_budget = torch.FloatTensor([self.target_budget
                                                ]).to(self.device)
        self.steepness = 10
        self.ceLoss = nn.CrossEntropyLoss()

        self.wandb_monitor = self.kwargs['CHIPNET_ARGS'].get('WANDB', 'False')
        self.dataset_name = dataloaders['train'].dataset.__class__.__name__
        self.save = './checkpoints/'

        self.name = '_'.join([self.dataset_name, 'ChipNet'])

        os.makedirs(f'{os.getcwd()}/logs/ChipNet', exist_ok=True)
        os.makedirs(self.save, exist_ok=True)

        if self.wandb_monitor:
            wandb.init(project='Trailmet ChipNet', name=self.name)
            wandb.config.update(self.kwargs)

    def compress_model(self):
        """Function to compress model using chipnet method."""
        self.model.to(self.device)

        if 'PRETRAIN' in self.kwargs:
            print('Pretrainning the model')
            self.kwargs['PRETRAIN']['phase'] = 'PRETRAIN'
            self.base_train(self.model, self.dataloaders,
                            **self.kwargs['PRETRAIN'])

        if 'PRUNE' in self.kwargs:
            print('preparing model for pruning')
            self.kwargs['PRUNE']['phase'] = 'PRUNE'
            self.prepare_model_for_compression()
            self.model.to(self.device)
            self.prune(self.model, self.dataloaders, **self.kwargs['PRUNE'])

        if 'FINETUNE' in self.kwargs:
            print('Finetuning the model')
            self.kwargs['FINETUNE']['phase'] = 'FINETUNE'
            self.prepare_for_finetuning(self.target_budget.item(),
                                        self.budget_type)
            self.base_train(self.model, self.dataloaders,
                            **self.kwargs['FINETUNE'])

    def base_train(self, model, dataloaders, **kwargs) -> None:
        """This function is used to perform standard model training and can be
        used for various purposes, such as model pretraining, fine-tuning of
        compressed models, among others.

        For cases, where base_train is not directly applicable, feel free to
        overwrite wherever this parent class is inherited.
        """
        best_top1_acc = 0  # setting to lowest possible value
        num_epochs = kwargs.get('EPOCHS', self.pretraining_epochs)
        test_only = kwargs.get('TEST_ONLY', False)

        ### preparing optimizer ###
        optimizer_name = kwargs.get('OPTIMIZER', 'SGD')
        lr = kwargs.get('LR', 0.05)
        scheduler_type = kwargs.get('SCHEDULER_TYPE', 1)
        weight_decay = kwargs.get('WEIGHT_DECAY', 0.001)

        optimizer = self.get_optimizer(optimizer_name, model, lr, weight_decay)

        criterion = nn.CrossEntropyLoss()

        phase = kwargs['phase']
        self.name = '_'.join([
            self.dataset_name,
            f'{num_epochs}',
            f'{lr}',
            datetime.now().strftime('%b-%d_%H:%M:%S'),
        ])

        self.logger_file = f'{os.getcwd()}/logs/ChipNet/{self.name}.log'

        logging.basicConfig(
            filename=self.logger_file,
            format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
        )

        logger.info(f'{phase} Arguments: {self.kwargs}')

        if test_only is False:
            epochs_list = []
            val_top1_acc_list = []
            val_top5_acc_list = []
            for epoch in range(num_epochs):
                adjust_learning_rate(optimizer, epoch, num_epochs,
                                     scheduler_type, lr)

                # t_loss = self.train_one_epoch(model, dataloaders['train'], criterion, optimizer)
                # acc, v_loss = self.test(model, dataloaders['val'], criterion)

                t_loss = self.train_one_epoch(
                    model,
                    dataloaders['train'],
                    criterion,
                    optimizer,
                    epoch,
                    phase,
                )

                valid_loss, valid_top1_acc, valid_top5_acc = self.test(
                    model, dataloaders['val'], criterion, epoch, phase)

                is_best = False
                if valid_top1_acc > best_top1_acc:
                    best_top1_acc = valid_top1_acc
                    is_best = True

                save_checkpoint(
                    {
                        'epoch': epoch,
                        'state_dict': model.state_dict(),
                        'best_top1_acc': best_top1_acc,
                        'optimizer': optimizer.state_dict(),
                    },
                    is_best,
                    self.save,
                )

                if self.wandb_monitor:
                    wandb.log({'best_top1_acc': best_top1_acc})

                epochs_list.append(epoch)
                val_top1_acc_list.append(valid_top1_acc.cpu().numpy())
                val_top5_acc_list.append(valid_top5_acc.cpu().numpy())

                df_data = np.array([
                    epochs_list,
                    val_top1_acc_list,
                    val_top5_acc_list,
                ]).T
                df = pd.DataFrame(
                    df_data,
                    columns=[
                        'Epochs',
                        'Validation Top1',
                        'Validation Top5',
                    ],
                )
                df.to_csv(
                    f'{os.getcwd()}/logs/ChipNet/{phase}_{self.name}.csv',
                    index=False,
                )

        state = torch.load(f'{self.save}/model_best.pth.tar')
        model.load_state_dict(state['state_dict'], strict=True)
        valid_loss, valid_top1_acc, valid_top5_acc = self.test(
            model, dataloaders['test'], criterion, epoch, phase)
        print(
            f"Test Accuracy: {valid_top1_acc} | Valid Accuracy: {state['best_top1_acc']}"
        )

    def train_one_epoch(
        self,
        model,
        dataloader,
        loss_fn,
        optimizer,
        epoch,
        phase,
        extra_functionality=None,
    ):
        """Standard training loop which can be used for various purposes with
        an extra functionality function to add to its working at the end of the
        loop."""
        model.train()

        batch_time = AverageMeter('Time', ':6.3f')
        data_time = AverageMeter('Data', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')

        end = time.time()

        epoch_iterator = tqdm(
            dataloader,
            desc=
            'Training X Epoch [X] (X / X Steps) (batch time=X.Xs) (data time=X.Xs) (loss=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        for i, (images, labels) in enumerate(epoch_iterator):
            data_time.update(time.time() - end)
            images = images.to(device=self.device)
            labels = labels.to(device=self.device)
            scores = model(images)

            loss = loss_fn(scores, labels)
            n = images.size(0)
            losses.update(loss.item(), n)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            batch_time.update(time.time() - end)
            end = time.time()

            epoch_iterator.set_description(
                'Training %s Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f)'
                % (
                    phase,
                    epoch,
                    (i + 1),
                    len(dataloader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                ))

            logger.info(
                'Training %s Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f)'
                % (
                    phase,
                    epoch,
                    (i + 1),
                    len(dataloader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                ))

            if self.wandb_monitor:
                wandb.log({
                    f'{phase}_train_loss': losses.val,
                })

            if extra_functionality is not None:
                extra_functionality()

        return losses.avg

    def test(self, model, dataloader, loss_fn=None, epoch=0, phase='PRETRAIN'):
        """This method is used to test the performance of the trained model."""

        model.eval()

        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            dataloader,
            desc=
            'Validating X Epoch [X] (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        with torch.no_grad():
            end = time.time()
            for i, (images, targets) in enumerate(epoch_iterator):
                images = images.to(self.device)
                targets = targets.to(self.device)
                outputs = model(images)
                pred1, pred5 = accuracy(outputs, targets, topk=(1, 5))

                if loss_fn is not None:
                    loss = loss_fn(outputs, targets)

                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Validating %s Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        phase,
                        epoch,
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Validating %s Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        phase,
                        epoch,
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        f'{phase}_val_loss': losses.val,
                        f'{phase}_val_top1_acc': top1.val,
                        f'{phase}_val_top5_acc': top5.val,
                    })

            print(' * acc@1 {top1.avg:.3f} acc@5 {top5.avg:.3f}'.format(
                top1=top1, top5=top5))

        return losses.avg, top1.avg, top5.avg

    def prune(self, model, dataloaders, **kwargs):
        """Function to prune a pretrained model using chipnet method."""
        num_epochs = kwargs.get('EPOCHS', 20)
        test_only = kwargs.get('TEST_ONLY', False)
        #### preparing optimizer ####
        lr = kwargs.get('LR', 0.001)
        weight_decay = kwargs.get('WEIGHT_DECAY', 0.001)
        param_optimizer = list(model.named_parameters())
        no_decay = ['zeta']
        optimizer_parameters = [
            {
                'params': [
                    p for n, p in param_optimizer
                    if not any(nd in n for nd in no_decay)
                ],
                'weight_decay':
                weight_decay,
                'lr':
                lr,
            },
            {
                'params': [
                    p for n, p in param_optimizer
                    if any(nd in n for nd in no_decay)
                ],
                'weight_decay':
                0.0,
                'lr':
                lr,
            },
        ]

        optimizer = optim.AdamW(optimizer_parameters)
        phase = kwargs['phase']
        criterion = self.prune_criterion
        best_acc = 0
        beta, gamma = 1.0, 2.0
        self.set_beta_gamma(beta, gamma)

        remaining_before_pruning = []
        remaining_after_pruning = []
        valid_accuracy = []
        pruning_accuracy = []
        pruning_threshold = []
        problems = []

        if test_only is False:
            for epoch in range(num_epochs):
                print(f'Starting epoch {epoch} / {num_epochs}')
                self.unprune_model()
                self.train_one_epoch(
                    model,
                    dataloaders['train'],
                    criterion,
                    optimizer,
                    epoch,
                    phase,
                    extra_functionality=self.steepness_update_function(
                        5.0 / len(dataloaders['train'])),
                )
                print(
                    f'[{epoch + 1} / {num_epochs}] Validation before pruning')
                logger.info(
                    f'[{epoch + 1} / {num_epochs}] Validation before pruning')
                _, acc, _ = self.test(model, dataloaders['val'], criterion,
                                      epoch, phase)
                remaining = self.get_remaining(self.steepness,
                                               self.budget_type).item()
                remaining_before_pruning.append(remaining)
                valid_accuracy.append(acc.cpu().numpy())

                print(f'[{epoch + 1} / {num_epochs}] Validation after pruning')
                logger.info(
                    f'[{epoch + 1} / {num_epochs}] Validation after pruning')
                threshold, problem = self.prune_model(self.target_budget,
                                                      self.budget_type)
                _, acc, _ = self.test(model, dataloaders['val'], criterion,
                                      epoch, phase)
                remaining = self.get_remaining(self.steepness,
                                               self.budget_type).item()
                pruning_accuracy.append(acc.cpu().numpy())
                pruning_threshold.append(threshold)
                remaining_after_pruning.append(remaining)
                problems.append(problem)

                beta = min(6.0, beta + (0.1 / self.b_inc))
                gamma = min(256, gamma * (2**(1.0 / self.g_inc)))
                self.set_beta_gamma(beta, gamma)
                print('Changed beta to', beta, 'changed gamma to', gamma)

                if acc > best_acc:
                    print('**Saving checkpoint**')
                    best_acc = acc
                    torch.save(
                        {
                            'epoch': epoch + 1,
                            'beta': beta,
                            'gamma': gamma,
                            'prune_threshold': threshold,
                            'state_dict': model.state_dict(),
                            'accuracy': acc,
                        },
                        f'checkpoints/{self.log_name}.pth',
                    )

                is_best = False
                if acc > best_acc:
                    best_acc = acc
                    is_best = True

                save_checkpoint(
                    {
                        'epoch': epoch,
                        'beta': beta,
                        'gamma': gamma,
                        'prune_threshold': threshold,
                        'state_dict': model.state_dict(),
                        'accuracy': acc,
                        'optimizer': optimizer.state_dict(),
                    },
                    is_best,
                    self.save,
                )

                df_data = np.array([
                    remaining_before_pruning,
                    remaining_after_pruning,
                    valid_accuracy,
                    pruning_accuracy,
                    pruning_threshold,
                    problems,
                ]).T
                df = pd.DataFrame(
                    df_data,
                    columns=[
                        'Remaining before pruning',
                        'Remaining after pruning',
                        'Valid accuracy',
                        'Pruning accuracy',
                        'Pruning threshold',
                        'problems',
                    ],
                )

                df.to_csv(
                    f'{os.getcwd()}/logs/ChipNet/{phase}_{self.name}.csv',
                    index=False,
                )

    def steepness_update_function(self, step):
        """Returns function to update steepness in budget loss of chipnet."""

        def update():
            self.steepness = min(
                60, self.steepness + step
            )  # increasing schedule of steepness to a maximum value of 60 to avoid gradient explosion.

        return update

    def prepare_model_for_compression(self):
        """Prepares model for compression by replacing batchnorm layers."""
        ModuleInjection.pruning_method = 'prune'

        def replace_bn(m):
            for attr_str in dir(m):
                target_attr = getattr(m, attr_str)
                if type(target_attr) == torch.nn.BatchNorm2d:
                    conv_attr = getattr(m, attr_str.replace('bn', 'conv'))
                    conv, bn = ModuleInjection.make_prunable(
                        conv_attr, target_attr)
                    setattr(m, attr_str.replace('bn', 'conv'), conv)
                    setattr(m, attr_str, bn)
            for _, ch in m.named_children():
                replace_bn(ch)

        self.prunable_modules = ModuleInjection.prunable_modules
        replace_bn(self.model)

    def prune_criterion(self, y_pred, y_true):
        """Loss function for pruning."""
        ce_loss = self.ceLoss(y_pred, y_true)
        budget_loss = ((self.get_remaining(self.steepness,
                                           self.budget_type).to(self.device) -
                        self.target_budget.to(self.device))**2).to(self.device)
        crispness_loss = self.get_crispnessLoss()
        return (budget_loss * self.budget_loss_weightage +
                crispness_loss * self.crispness_loss_weightage + ce_loss)

    def calculate_prune_threshold(self, target_budget, budget_type):
        """Calculates the prune threshold for different budget types."""
        zetas = self.give_zetas()
        if budget_type in ['volume_ratio']:
            zeta_weights = self.give_zeta_weights()
            zeta_weights = zeta_weights[np.argsort(zetas)]
        zetas = sorted(zetas)
        if budget_type == 'volume_ratio':
            curr_budget = 0
            indx = 0
            while curr_budget < (1.0 - target_budget):
                indx += 1
                curr_budget += zeta_weights[indx]
            prune_threshold = zetas[indx]
        else:
            prune_threshold = zetas[int((1.0 - target_budget) * len(zetas))]
        return prune_threshold

    def smoothRound(self, x, steepness=20.0):
        """Function to apply smooth rounding on zeta for more accurate budget
        calculation."""
        return 1.0 / (1.0 + torch.exp(-1 * steepness * (x - 0.5)))

    def n_remaining(self, module, steepness=20.0):
        """Returns the remaining number of channels."""
        return (module.pruned_zeta if module.is_pruned else self.smoothRound(
            module.get_zeta_t(), steepness)).sum()

    def is_all_pruned(self, module):
        """Checks if the whole block is pruned."""
        return self.n_remaining(module) == 0

    def get_remaining(self, steepness=20.0, budget_type='channel_ratio'):
        """Return the fraction of active zeta_t (i.e > 0.5)"""
        n_rem = 0
        n_total = 0
        for l_block in self.prunable_modules:
            if budget_type == 'volume_ratio':
                n_rem += (self.n_remaining(l_block, steepness) *
                          l_block._conv_module.output_area)
                n_total += l_block.num_gates * l_block._conv_module.output_area
            elif budget_type == 'channel_ratio':
                n_rem += self.n_remaining(l_block, steepness)
                n_total += l_block.num_gates
            elif budget_type == 'parameter_ratio':
                k = l_block._conv_module.kernel_size[0]
                prev_total = (3 if self.prev_module[l_block] is None else
                              self.prev_module[l_block].num_gates)
                prev_remaining = (3 if self.prev_module[l_block] is None
                                  else self.n_remaining(
                                      self.prev_module[l_block], steepness))
                n_rem += self.n_remaining(l_block,
                                          steepness) * prev_remaining * k * k
                n_total += l_block.num_gates * prev_total * k * k
            elif budget_type == 'flops_ratio':
                k = l_block._conv_module.kernel_size[0]
                output_area = l_block._conv_module.output_area
                prev_total = (3 if self.prev_module[l_block] is None else
                              self.prev_module[l_block].num_gates)
                prev_remaining = (3 if self.prev_module[l_block] is None
                                  else self.n_remaining(
                                      self.prev_module[l_block], steepness))
                curr_remaining = self.n_remaining(l_block, steepness)
                n_rem += (
                    curr_remaining * prev_remaining * k * k * output_area +
                    curr_remaining * output_area)
                n_total += (
                    l_block.num_gates * prev_total * k * k * output_area +
                    l_block.num_gates * output_area)
        return n_rem / n_total

    def give_zetas(self):
        """Returns pruning gates as a list."""
        zetas = []
        for l_block in self.prunable_modules:
            zetas.append(l_block.get_zeta_t().cpu().detach().numpy().tolist())
        zetas = [z for k in zetas for z in k]
        return zetas

    def give_zeta_weights(self):
        """Returns the importance of pruning gates using the volume it
        represents, used for volume pruning."""
        zeta_weights = []
        for l_block in self.prunable_modules:
            zeta_weights.append([l_block._conv_module.output_area] *
                                l_block.num_gates)
        zeta_weights = [z for k in zeta_weights for z in k]
        return zeta_weights / np.sum(zeta_weights)

    def plot_zt(self):
        """Plots the distribution of zeta_t and returns the same."""
        zetas = self.give_zetas()
        exactly_zeros = np.sum(np.array(zetas) == 0.0)
        exactly_ones = np.sum(np.array(zetas) == 1.0)
        plt.hist(zetas)
        plt.show()
        return exactly_zeros, exactly_ones

    def get_crispnessLoss(self):
        """Loss responsible for making zeta_t 1 or 0."""
        loss = torch.FloatTensor([]).to(self.device)
        for l_block in self.prunable_modules:
            loss = torch.cat([
                loss,
                torch.pow(l_block.get_zeta_t() - l_block.get_zeta_i(), 2)
            ])
        return torch.mean(loss).to(self.device)

    def prune_model(
        self,
        target_budget,
        budget_type='channel_ratio',
        finetuning=False,
        threshold=None,
    ):
        """Prunes the network to make zeta_t exactly 1 and 0."""

        if budget_type == 'parameter_ratio':
            zetas = sorted(self.give_zetas())
            high = len(zetas) - 1
            low = 0
            while low < high:
                mid = (high + low) // 2
                threshold = zetas[mid]
                for l_block in self.prunable_modules:
                    l_block.prune(threshold)
                self.remove_orphans()
                if self.params() < target_budget:
                    high = mid - 1
                else:
                    low = mid + 1
        elif budget_type == 'flops_ratio':
            zetas = sorted(self.give_zetas())
            high = len(zetas) - 1
            low = 0
            while low < high:
                mid = (high + low) // 2
                threshold = zetas[mid]
                for l_block in self.prunable_modules:
                    l_block.prune(threshold)
                self.remove_orphans()
                if self.flops() < target_budget:
                    high = mid - 1
                else:
                    low = mid + 1
        else:
            if threshold == None:
                self.prune_threshold = self.calculate_prune_threshold(
                    target_budget, budget_type)
                threshold = min(self.prune_threshold, 0.9)

        for l_block in self.prunable_modules:
            l_block.prune(threshold)

        if finetuning:
            self.remove_orphans()
            return threshold
        else:
            problem = self.check_abnormality()
            return threshold, problem

    def unprune_model(self):
        """Unprunes the network to again make pruning gates continuous."""
        for l_block in self.prunable_modules:
            l_block.unprune()

    def prepare_for_finetuning(self, budget, budget_type='channel_ratio'):
        """Freezes zeta."""
        self.model(torch.rand(2, 3, 32, 32).to(self.device))
        threshold = self.prune_model(budget,
                                     budget_type=budget_type,
                                     finetuning=True)
        if budget_type not in ['parameter_ratio', 'flops_ratio']:
            while self.get_remaining(steepness=20.0,
                                     budget_type=budget_type) < budget:
                threshold -= 0.0001
                self.prune_model(
                    budget,
                    finetuning=True,
                    budget_type=budget_type,
                    threshold=threshold,
                )
        return threshold

    def get_params_count(self):
        """Returns the number of active and total parameters in the network."""
        total_params = 0
        active_params = 0
        for l_block in self.model.modules():
            if isinstance(l_block, PrunableBatchNorm2d):
                active_param, total_param = l_block.get_params_count()
                active_params += active_param
                total_params += total_param
            if isinstance(l_block, nn.Linear):
                linear_params = l_block.weight.view(-1).shape[0]
                active_params += linear_params
                total_params += linear_params
        return active_params, total_params

    def get_volume(self):
        """Returns the active and total volume of the network."""
        total_volume = 0.0
        active_volume = 0.0
        for l_block in self.prunable_modules:
            active_volume_, total_volume_ = l_block.get_volume()
            active_volume += active_volume_
            total_volume += total_volume_
        return active_volume, total_volume

    def get_flops(self):
        """Returns the active and total flops of the network."""
        total_flops = 0.0
        active_flops = 0.0
        for l_block in self.prunable_modules:
            active_flops_, total_flops_ = l_block.get_flops()
            active_flops += active_flops_
            total_flops += total_flops_
        return active_flops, total_flops

    def get_channels(self):
        """Returns the active and total number of channels in the network."""
        total_channels = 0.0
        active_channels = 0.0
        for l_block in self.prunable_modules:
            active_channels += l_block.pruned_zeta.sum().item()
            total_channels += l_block.num_gates
        return active_channels, total_channels

    def set_beta_gamma(self, beta, gamma):
        """Explicitly sets the values of beta and gamma parameters."""
        for l_block in self.prunable_modules:
            l_block.set_beta_gamma(beta, gamma)

    def check_abnormality(self):
        """Checks for any abnormality in the pruning process."""
        n_removable = self.removable_orphans()
        isbroken = self.check_if_broken()
        if n_removable != 0.0 and isbroken:
            return f'both rem_{n_removable} and broken'
        if n_removable != 0.0:
            return f'removable_{n_removable}'
        if isbroken:
            return 'broken'

    def check_if_broken(self):
        """Checks if the network is broken due to abnormal pruning."""
        for bn in self.prunable_modules:
            if bn.is_imp and bn.pruned_zeta.sum() == 0:
                return True
        return False

    def removable_orphans(self):
        """Checks for orphan nodes from the network that are not being used in
        the computation graph."""
        num_removed = 0
        bn_layers = self.model.get_bn_layers()
        for l_blocks in bn_layers:
            flag = sum([self.is_all_pruned(m) for m in l_blocks])
            if flag:
                num_removed += sum([self.n_remaining(m) for m in l_blocks])
        return num_removed

    def remove_orphans(self):
        """Removes orphan nodes from the network that are not being used in the
        computation graph."""
        num_removed = 0
        bn_layers = self.model.get_bn_layers()
        for l_blocks in bn_layers:
            flag = sum([self.is_all_pruned(m) for m in l_blocks])
            if flag:
                num_removed += sum([self.n_remaining(m) for m in l_blocks])
                for m in l_blocks:
                    m.pruned_zeta.data.copy_(torch.zeros_like(m.pruned_zeta))
        return num_removed



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Union
from trailmet.models.resnet import BasicBlock, Bottleneck
from trailmet.models.mobilenet import InvertedResidual
from trailmet.algorithms.quantize.quantize import StraightThrough
from trailmet.algorithms.quantize.methods import (
    MaxAbsStaticQuantization,
    LpNormQuantization,
)
from trailmet.algorithms.quantize.methods import UniformAffineQuantizer
from trailmet.algorithms.quantize.methods import ActQuantizer

__all__ = [
    'QuantBasicBlock',
    'QuantBottleneck',
    'QuantInvertedResidual',
    'QuantModule',
    'BaseQuantBlock',
    'QBasicBlock',
    'QBottleneck',
    'QInvertedResidual',
    'ActivationModuleWrapper',
    'ParameterModuleWrapper',
]
# ============================================
# ***** Quantization Modules for BRECQ *******
# ============================================
"""
Supported quantization wrappers for pytorch modules :-
    - BasicBlock(nn.Module) -> QuantBasicBlock(BaseQuantBlock(nn.Module))
    - Bottleneck(nn.Module) -> QuantBottleneck(BaseQuantBlock(nn.Module))
    - InvertedResidual(nn.Module) -> QuantInvertedResidual(BaseQuantBlock(nn.Module))
        - nn.Conv2d, nn.Linear -> QuantModule(nn.Module)
"""


class QuantModule(nn.Module):
    """Quantized Module that can perform quantized convolution or normal
    convolution.

    To activate quantization, please use set_quant_state function.
    """

    def __init__(
        self,
        org_module: Union[nn.Conv2d, nn.Linear],
        weight_quant_params: dict = {},
        act_quant_params: dict = {},
        disable_act_quant: bool = False,
        se_module=None,
    ):
        super(QuantModule, self).__init__()
        if isinstance(org_module, nn.Conv2d):
            self.fwd_kwargs = dict(
                stride=org_module.stride,
                padding=org_module.padding,
                dilation=org_module.dilation,
                groups=org_module.groups,
            )
            self.fwd_func = F.conv2d
        else:
            self.fwd_kwargs = dict()
            self.fwd_func = F.linear
        self.weight = org_module.weight
        self.org_weight = org_module.weight.data.clone()
        if org_module.bias is not None:
            self.bias = org_module.bias
            self.org_bias = org_module.bias.data.clone()
        else:
            self.bias = None
            self.org_bias = None
        # de-activate the quantized forward default
        self.use_weight_quant = False
        self.use_act_quant = False
        self.disable_act_quant = disable_act_quant
        # initialize quantizer
        self.weight_quantizer = UniformAffineQuantizer(**weight_quant_params)
        self.act_quantizer = UniformAffineQuantizer(**act_quant_params)

        self.activation_function = StraightThrough()
        self.ignore_reconstruction = False

        self.se_module = se_module
        self.extra_repr = org_module.extra_repr

    def forward(self, input: torch.Tensor):
        if self.use_weight_quant:
            weight = self.weight_quantizer(self.weight)
            bias = self.bias
        else:
            weight = self.org_weight
            bias = self.org_bias
        out = self.fwd_func(input, weight, bias, **self.fwd_kwargs)
        # disable act quantization is designed for convolution before elemental-wise operation,
        # in that case, we apply activation function and quantization after ele-wise op.
        if self.se_module is not None:
            out = self.se_module(out)
        out = self.activation_function(out)
        if self.disable_act_quant:
            return out
        if self.use_act_quant:
            out = self.act_quantizer(out)
        return out

    def set_quant_state(self,
                        weight_quant: bool = False,
                        act_quant: bool = False):
        self.use_weight_quant = weight_quant
        self.use_act_quant = act_quant


class BaseQuantBlock(nn.Module):
    """Base implementation of block structures for all networks.

    Due to the branch architecture, we have to perform activation function and
    quantization after the elemental-wise add operation, therefore, we put this
    part in this class.
    """

    def __init__(self, act_quant_params: dict = {}):
        super().__init__()
        self.use_weight_quant = False
        self.use_act_quant = False
        # initialize quantizer

        self.act_quantizer = UniformAffineQuantizer(**act_quant_params)
        self.activation_function = StraightThrough()

        self.ignore_reconstruction = False

    def set_quant_state(self,
                        weight_quant: bool = False,
                        act_quant: bool = False):
        # setting weight quantization here does not affect actual forward pass
        self.use_weight_quant = weight_quant
        self.use_act_quant = act_quant
        for m in self.modules():
            if isinstance(m, QuantModule):
                m.set_quant_state(weight_quant, act_quant)


class QuantBasicBlock(BaseQuantBlock):
    """Implementation of Quantized BasicBlock used in ResNet-18 and
    ResNet-34."""

    def __init__(
        self,
        basic_block: BasicBlock,
        weight_quant_params: dict = {},
        act_quant_params: dict = {},
    ):
        super().__init__(act_quant_params)
        self.conv1 = QuantModule(basic_block.conv1, weight_quant_params,
                                 act_quant_params)
        self.conv1.activation_function = basic_block.active
        self.conv2 = QuantModule(
            basic_block.conv2,
            weight_quant_params,
            act_quant_params,
            disable_act_quant=True,
        )

        # modify the activation function to ReLU
        self.activation_function = basic_block.active

        if basic_block.downsample is None:
            self.downsample = None
        else:
            self.downsample = QuantModule(
                basic_block.downsample[0],
                weight_quant_params,
                act_quant_params,
                disable_act_quant=True,
            )
        # copying all attributes in original block
        self.stride = basic_block.stride

    def forward(self, x):
        residual = x if self.downsample is None else self.downsample(x)
        out = self.conv1(x)
        out = self.conv2(out)
        out += residual
        out = self.activation_function(out)
        if self.use_act_quant:
            out = self.act_quantizer(out)
        return out


class QuantBottleneck(BaseQuantBlock):
    """Implementation of Quantized Bottleneck Block used in ResNet-50, -101 and
    -152."""

    def __init__(
        self,
        bottleneck: Bottleneck,
        weight_quant_params: dict = {},
        act_quant_params: dict = {},
    ):
        super().__init__(act_quant_params)
        self.conv1 = QuantModule(bottleneck.conv1, weight_quant_params,
                                 act_quant_params)
        self.conv1.activation_function = bottleneck.active
        self.conv2 = QuantModule(bottleneck.conv2, weight_quant_params,
                                 act_quant_params)
        self.conv2.activation_function = bottleneck.active
        self.conv3 = QuantModule(
            bottleneck.conv3,
            weight_quant_params,
            act_quant_params,
            disable_act_quant=True,
        )

        # modify the activation function to ReLU
        self.activation_function = bottleneck.active

        if bottleneck.downsample is None:
            self.downsample = None
        else:
            self.downsample = QuantModule(
                bottleneck.downsample[0],
                weight_quant_params,
                act_quant_params,
                disable_act_quant=True,
            )
        # copying all attributes in original block
        self.stride = bottleneck.stride

    def forward(self, x):
        residual = x if self.downsample is None else self.downsample(x)
        out = self.conv1(x)
        out = self.conv2(out)
        out = self.conv3(out)
        out += residual
        out = self.activation_function(out)
        if self.use_act_quant:
            out = self.act_quantizer(out)
        return out


class QuantInvertedResidual(BaseQuantBlock):
    """Implementation of Quantized Inverted Residual Block used in MobileNetV2.

    Inverted Residual does not have activation function.
    """

    def __init__(
        self,
        inv_res: InvertedResidual,
        weight_quant_params: dict = {},
        act_quant_params: dict = {},
    ):
        super().__init__(act_quant_params)
        self.stride = inv_res.stride
        self.inp = inv_res.inp
        self.oup = inv_res.oup
        self.exp = inv_res.exp
        self.conv1 = QuantModule(inv_res.conv1, weight_quant_params,
                                 act_quant_params)
        self.conv1.activation_function = nn.ReLU6(inplace=True)
        self.conv2 = QuantModule(inv_res.conv2, weight_quant_params,
                                 act_quant_params)
        self.conv2.activation_function = nn.ReLU6(inplace=True)
        self.conv3 = QuantModule(inv_res.conv3, weight_quant_params,
                                 act_quant_params)
        self.shortcut = nn.Sequential()
        if self.stride == 1 and self.inp != self.oup:
            self.shortcut = nn.Sequential(
                QuantModule(inv_res.shortcut[0], weight_quant_params,
                            act_quant_params))
        # self.use_res_connect = inv_res.use_res_connect
        # self.expand_ratio = inv_res.exp
        # if self.expand_ratio == 1:
        #     self.conv = nn.Sequential(
        #         QuantModule(inv_res.conv[0], weight_quant_params, act_quant_params),
        #         QuantModule(inv_res.conv[3], weight_quant_params, act_quant_params, disable_act_quant=True),
        #     )
        #     self.conv[0].activation_function = nn.ReLU6()
        # else:
        #     self.conv = nn.Sequential(
        #         QuantModule(inv_res.conv[0], weight_quant_params, act_quant_params),
        #         QuantModule(inv_res.conv[3], weight_quant_params, act_quant_params),
        #         QuantModule(inv_res.conv[6], weight_quant_params, act_quant_params, disable_act_quant=True),
        #     )
        #     self.conv[0].activation_function = nn.ReLU6()
        #     self.conv[1].activation_function = nn.ReLU6()

    def forward(self, x):
        out = self.conv1(x)
        out = self.conv2(out)
        out = self.conv3(out)
        out = out + self.shortcut(x) if self.stride == 1 else out
        return out
        # if self.use_res_connect:
        #     out = x + self.conv(x)
        # else:
        #     out = self.conv(x)
        # out = self.activation_function(out)
        # if self.use_act_quant:
        #     out = self.act_quantizer(out)
        # return out


# ===============================================
# ***** Quantization Modules for BitSplit *******
# ===============================================
"""
Supported quantization wrappers for pytorch modules :-
    - BasicBlock(nn.Module) -> QBasicBlock(nn.Module)
    - Bottleneck(nn.Module) -> QBottleneck(nn.Module)
    - InvertedResidual(nn.Module) -> QInvertedResidual(nn.Module)
"""


class QBasicBlock(nn.Module):
    expansion = 1

    def __init__(self, basic_block: BasicBlock):
        super().__init__()
        self.quant1 = ActQuantizer()
        self.conv1 = basic_block.conv1
        self.bn1 = basic_block.bn1
        self.active = basic_block.active
        self.quant2 = ActQuantizer()
        self.conv2 = basic_block.conv2
        self.bn2 = basic_block.bn2
        self.downsample = basic_block.downsample
        self.stride = basic_block.stride

    def forward(self, x):
        residual = x
        x = self.quant1(x)
        out = self.active(self.bn1(self.conv1(x)))
        out = self.quant2(out)
        out = self.bn2(self.conv2(out))
        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual
        out = self.active(out)
        return out


class QBottleneck(nn.Module):
    expansion = 4

    def __init__(self, bottleneck: Bottleneck):
        super().__init__()
        self.quant1 = ActQuantizer()
        self.conv1 = bottleneck.conv1
        self.bn1 = bottleneck.bn1
        self.quant2 = ActQuantizer()
        self.conv2 = bottleneck.conv2
        self.bn2 = bottleneck.bn2
        self.quant3 = ActQuantizer()
        self.conv3 = bottleneck.conv3
        self.bn3 = bottleneck.bn3
        self.active = bottleneck.active
        self.downsample = bottleneck.downsample
        self.stride = bottleneck.stride

    def forward(self, x):
        residual = x
        x = self.quant1(x)
        out = self.active(self.bn1(self.conv1(x)))
        out = self.quant2(out)
        out = self.active(self.bn2(self.conv2(out)))
        out = self.quant3(out)
        out = self.bn3(self.conv3(out))
        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual
        out = self.active(out)
        return out


class QInvertedResidual(nn.Module):

    def __init__(self, inv_res: InvertedResidual):
        super().__init__()
        self.stride = inv_res.stride
        self.inp = inv_res.inp
        self.oup = inv_res.oup
        self.exp = inv_res.exp
        self.quant1 = ActQuantizer(islinear=1)
        self.conv1 = inv_res.conv1
        self.bn1 = inv_res.bn1
        self.quant2 = ActQuantizer(islinear=1)
        self.conv2 = inv_res.conv2
        self.bn2 = inv_res.bn2
        self.quant3 = ActQuantizer(islinear=0)
        self.conv3 = inv_res.conv3
        self.bn3 = inv_res.bn3
        self.shortcut = inv_res.shortcut

    def forward(self, x):
        x = self.quant1(x)
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.quant2(out)
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.quant3(out)
        out = self.bn3(self.conv3(out))
        out = out + self.shortcut(x) if self.stride == 1 else out
        return out


# ===========================================
# ***** Quantization Modules for LAPQ *******
# ===========================================
"""
Supported quantization wrappers for pytorch modules :-
    - nn.ReLU, nn.ReLU6 -> ActivationModuleWrapper(nn.Module)
    - nn.Conv2d, nn.Linear -> ParameterModuleWrapper(nn.Module)
"""

quantization_mapping = {
    'max_static': MaxAbsStaticQuantization,
    'lp_norm': LpNormQuantization,
}


def is_positive(module):
    return isinstance(module, nn.ReLU) or isinstance(module, nn.ReLU6)


class ActivationModuleWrapper(nn.Module):

    def __init__(self, name, wrapped_module, **kwargs):
        super(ActivationModuleWrapper, self).__init__()
        self.name = name
        self.wrapped_module = wrapped_module
        self.bits_out = kwargs['bits_out']
        self.qtype = kwargs['qtype']
        self.post_relu = True
        self.enabled = True
        self.active = True
        if self.bits_out is not None:
            self.out_quantization = self.out_quantization_default = None

            def __init_out_quantization__(tensor):
                self.out_quantization_default = quantization_mapping[
                    self.qtype](
                        self,
                        tensor,
                        self.bits_out,
                        symmetric=(not is_positive(wrapped_module)),
                        uint=True,
                        kwargs=kwargs,
                    )
                self.out_quantization = self.out_quantization_default

            self.out_quantization_init_fn = __init_out_quantization__

    def __enabled__(self):
        return self.enabled and self.active and self.bits_out is not None

    def forward(self, *input):
        if self.post_relu:
            out = self.wrapped_module(*input)
            # Quantize output
            if self.__enabled__():
                self.verify_initialized(self.out_quantization, out,
                                        self.out_quantization_init_fn)
                out = self.out_quantization(out)
        else:
            # Quantize output
            if self.__enabled__():
                self.verify_initialized(self.out_quantization, *input,
                                        self.out_quantization_init_fn)
                out = self.out_quantization(*input)
            else:
                out = self.wrapped_module(*input)
        return out

    @staticmethod
    def verify_initialized(quantization_handle, tensor, init_fn):
        if quantization_handle is None:
            init_fn(tensor)

    def get_quantization(self):
        return self.out_quantization

    def set_quantization(self, qtype, kwargs):
        self.out_quantization = qtype(
            self,
            self.bits_out,
            symmetric=(not is_positive(self.wrapped_module)),
            uint=True,
            kwargs=kwargs,
        )


class ParameterModuleWrapper(nn.Module):

    def __init__(self, name, wrapped_module, **kwargs):
        super(ParameterModuleWrapper, self).__init__()
        self.name = name
        self.wrapped_module = wrapped_module
        self.forward_functor = kwargs['forward_functor']
        self.bit_weights = kwargs['bit_weights']
        self.bits_out = kwargs['bits_out']
        self.qtype = kwargs['qtype']
        self.bcorr_w = kwargs['bcorr_w']
        self.bn = kwargs['bn'] if 'bn' in kwargs else None
        self.enabled = True
        self.active = True
        self.centroids_hist = {}
        self.log_weights_hist = False
        self.log_weights_mse = False
        self.log_clustering = False
        self.dynamic_weight_quantization = True
        setattr(self, 'weight', wrapped_module.weight)
        delattr(wrapped_module, 'weight')
        if hasattr(wrapped_module, 'bias'):
            setattr(self, 'bias', wrapped_module.bias)
            delattr(wrapped_module, 'bias')
        if self.bit_weights is not None:
            self.weight_quantization_default = quantization_mapping[
                self.qtype](
                    self,
                    self.weight,
                    self.bit_weights,
                    symmetric=True,
                    uint=True,
                    kwargs=kwargs,
                )
            self.weight_quantization = self.weight_quantization_default
            if not self.dynamic_weight_quantization:
                self.weight_q = self.weight_quantization(self.weight)
                self.weight_mse = torch.mean(
                    (self.weight_q - self.weight)**2).item()

    def __enabled__(self):
        return self.enabled and self.active and self.bit_weights is not None

    def bias_corr(self, x, xq):
        bias_q = xq.view(xq.shape[0], -1).mean(-1)
        bias_orig = x.view(x.shape[0], -1).mean(-1)
        bcorr = bias_q - bias_orig
        return (xq - bcorr.view(bcorr.numel(), 1, 1, 1)
                if len(x.shape) == 4 else xq - bcorr.view(bcorr.numel(), 1))

    def forward(self, *input):
        w = self.weight
        if self.__enabled__():
            # Quantize weights
            if self.dynamic_weight_quantization:
                w = self.weight_quantization(self.weight)
                if self.bcorr_w:
                    w = self.bias_corr(self.weight, w)
            else:
                w = self.weight_q
        out = self.forward_functor(
            *input,
            weight=w,
            bias=(self.bias if hasattr(self, 'bias') else None))
        return out

    def get_quantization(self):
        return self.weight_quantization

    def set_quantization(self, qtype, kwargs):
        self.weight_quantization = qtype(self,
                                         self.bit_weights,
                                         symmetric=True,
                                         uint=True,
                                         kwargs=kwargs)
